@@ -23,7 +23,6 @@
  */
 
 #define __USE_GNU
-#define _GNU_SOURCE
 
 #include <limits.h>
 #ifndef __FreeBSD__
@@ -60,6 +59,7 @@ typedef struct {
 
 	char *node_name;
 	char *target;
+	char *role;
 
 	int fd;
 	int error;
@@ -500,6 +500,8 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 
 	if (pw_properties_get(props, PW_KEY_NODE_LATENCY) == NULL)
 		pw_properties_setf(props, PW_KEY_NODE_LATENCY, "%lu/%u", pw->min_avail, io->rate);
+	if (pw_properties_get(props, PW_KEY_NODE_RATE) == NULL)
+		pw_properties_setf(props, PW_KEY_NODE_RATE, "1/%u", io->rate);
 	if (pw->target != NULL &&
 		pw_properties_get(props, PW_KEY_NODE_TARGET) == NULL)
 		pw_properties_setf(props, PW_KEY_NODE_TARGET, "%s", pw->target);
@@ -510,6 +512,9 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 		pw_properties_set(props, PW_KEY_MEDIA_CATEGORY,
 				io->stream == SND_PCM_STREAM_PLAYBACK ?
 				"Playback" : "Capture");
+	if (pw->role != NULL &&
+		pw_properties_get(props, PW_KEY_MEDIA_ROLE) == NULL)
+		pw_properties_setf(props, PW_KEY_MEDIA_ROLE, "%s", pw->role);
 
 	pw->stream = pw_stream_new(pw->core, pw->node_name, props);
 	if (pw->stream == NULL)
@@ -959,6 +964,7 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp, const char *name,
 				const char *server_name,
 				const char *playback_node,
 				const char *capture_node,
+				const char *role,
 				snd_pcm_stream_t stream,
 				int mode,
 				uint32_t flags,
@@ -1014,6 +1020,8 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp, const char *name,
 			pw->target = capture_node ? strdup(capture_node) : NULL;
 	}
 
+	pw->role = (role && *role) ? strdup(role) : NULL;
+
 	pw->main_loop = pw_thread_loop_new("alsa-pipewire", NULL);
 	if (pw->main_loop == NULL) {
 		err = -errno;
@@ -1021,7 +1029,11 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp, const char *name,
 	}
 	loop = pw_thread_loop_get_loop(pw->main_loop);
 	pw->system = loop->system;
-	if ((pw->context = pw_context_new(loop, NULL, 0)) == NULL) {
+	if ((pw->context = pw_context_new(loop,
+					pw_properties_new(
+						PW_KEY_CONFIG_NAME, "client-rt.conf",
+						NULL),
+					0)) == NULL) {
 		err = -errno;
 		goto error;
 	}
@@ -1092,6 +1104,7 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 	const char *server_name = NULL;
 	const char *playback_node = NULL;
 	const char *capture_node = NULL;
+	const char *role = NULL;
 	snd_pcm_format_t format = SND_PCM_FORMAT_UNKNOWN;
 	int rate = 0;
 	int channels = 0;
@@ -1126,6 +1139,10 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 			snd_config_get_string(n, &capture_node);
 			continue;
 		}
+		if (spa_streq(id, "role")) {
+			snd_config_get_string(n, &role);
+			continue;
+		}
 		if (spa_streq(id, "exclusive")) {
 			if (snd_config_get_bool(n))
 				flags |= PW_STREAM_FLAG_EXCLUSIVE;
@@ -1175,7 +1192,7 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 	}
 
 	err = snd_pcm_pipewire_open(pcmp, name, node_name, server_name, playback_node,
-			capture_node, stream, mode, flags, rate, format,
+			capture_node, role, stream, mode, flags, rate, format,
 			channels, period_bytes);
 
 	return err;
