@@ -26,6 +26,8 @@
 #include "alsa-mixer.h"
 #include "alsa-ucm.h"
 
+#include <spa/utils/string.h>
+
 int _acp_log_level = 1;
 acp_log_func _acp_log_func;
 void *_acp_log_data;
@@ -52,38 +54,38 @@ static const uint32_t channel_table[PA_CHANNEL_POSITION_MAX] = {
 	[PA_CHANNEL_POSITION_SIDE_LEFT] = ACP_CHANNEL_SL,
 	[PA_CHANNEL_POSITION_SIDE_RIGHT] = ACP_CHANNEL_SR,
 
-	[PA_CHANNEL_POSITION_AUX0] = ACP_CHANNEL_CUSTOM_START + 1,
-	[PA_CHANNEL_POSITION_AUX1] = ACP_CHANNEL_CUSTOM_START + 2,
-	[PA_CHANNEL_POSITION_AUX2] = ACP_CHANNEL_CUSTOM_START + 3,
-	[PA_CHANNEL_POSITION_AUX3] = ACP_CHANNEL_CUSTOM_START + 4,
-	[PA_CHANNEL_POSITION_AUX4] = ACP_CHANNEL_CUSTOM_START + 5,
-	[PA_CHANNEL_POSITION_AUX5] = ACP_CHANNEL_CUSTOM_START + 6,
-	[PA_CHANNEL_POSITION_AUX6] = ACP_CHANNEL_CUSTOM_START + 7,
-	[PA_CHANNEL_POSITION_AUX7] = ACP_CHANNEL_CUSTOM_START + 8,
-	[PA_CHANNEL_POSITION_AUX8] = ACP_CHANNEL_CUSTOM_START + 9,
-	[PA_CHANNEL_POSITION_AUX9] = ACP_CHANNEL_CUSTOM_START + 10,
-	[PA_CHANNEL_POSITION_AUX10] = ACP_CHANNEL_CUSTOM_START + 11,
-	[PA_CHANNEL_POSITION_AUX11] = ACP_CHANNEL_CUSTOM_START + 12,
-	[PA_CHANNEL_POSITION_AUX12] = ACP_CHANNEL_CUSTOM_START + 13,
-	[PA_CHANNEL_POSITION_AUX13] = ACP_CHANNEL_CUSTOM_START + 14,
-	[PA_CHANNEL_POSITION_AUX14] = ACP_CHANNEL_CUSTOM_START + 15,
-	[PA_CHANNEL_POSITION_AUX15] = ACP_CHANNEL_CUSTOM_START + 16,
-	[PA_CHANNEL_POSITION_AUX16] = ACP_CHANNEL_CUSTOM_START + 17,
-	[PA_CHANNEL_POSITION_AUX17] = ACP_CHANNEL_CUSTOM_START + 18,
-	[PA_CHANNEL_POSITION_AUX18] = ACP_CHANNEL_CUSTOM_START + 19,
-	[PA_CHANNEL_POSITION_AUX19] = ACP_CHANNEL_CUSTOM_START + 20,
-	[PA_CHANNEL_POSITION_AUX20] = ACP_CHANNEL_CUSTOM_START + 21,
-	[PA_CHANNEL_POSITION_AUX21] = ACP_CHANNEL_CUSTOM_START + 22,
-	[PA_CHANNEL_POSITION_AUX22] = ACP_CHANNEL_CUSTOM_START + 23,
-	[PA_CHANNEL_POSITION_AUX23] = ACP_CHANNEL_CUSTOM_START + 24,
-	[PA_CHANNEL_POSITION_AUX24] = ACP_CHANNEL_CUSTOM_START + 25,
-	[PA_CHANNEL_POSITION_AUX25] = ACP_CHANNEL_CUSTOM_START + 26,
-	[PA_CHANNEL_POSITION_AUX26] = ACP_CHANNEL_CUSTOM_START + 27,
-	[PA_CHANNEL_POSITION_AUX27] = ACP_CHANNEL_CUSTOM_START + 28,
-	[PA_CHANNEL_POSITION_AUX28] = ACP_CHANNEL_CUSTOM_START + 29,
-	[PA_CHANNEL_POSITION_AUX29] = ACP_CHANNEL_CUSTOM_START + 30,
-	[PA_CHANNEL_POSITION_AUX30] = ACP_CHANNEL_CUSTOM_START + 31,
-	[PA_CHANNEL_POSITION_AUX31] = ACP_CHANNEL_CUSTOM_START + 32,
+	[PA_CHANNEL_POSITION_AUX0] = ACP_CHANNEL_START_Aux + 0,
+	[PA_CHANNEL_POSITION_AUX1] = ACP_CHANNEL_START_Aux + 1,
+	[PA_CHANNEL_POSITION_AUX2] = ACP_CHANNEL_START_Aux + 2,
+	[PA_CHANNEL_POSITION_AUX3] = ACP_CHANNEL_START_Aux + 3,
+	[PA_CHANNEL_POSITION_AUX4] = ACP_CHANNEL_START_Aux + 4,
+	[PA_CHANNEL_POSITION_AUX5] = ACP_CHANNEL_START_Aux + 5,
+	[PA_CHANNEL_POSITION_AUX6] = ACP_CHANNEL_START_Aux + 6,
+	[PA_CHANNEL_POSITION_AUX7] = ACP_CHANNEL_START_Aux + 7,
+	[PA_CHANNEL_POSITION_AUX8] = ACP_CHANNEL_START_Aux + 8,
+	[PA_CHANNEL_POSITION_AUX9] = ACP_CHANNEL_START_Aux + 9,
+	[PA_CHANNEL_POSITION_AUX10] = ACP_CHANNEL_START_Aux + 10,
+	[PA_CHANNEL_POSITION_AUX11] = ACP_CHANNEL_START_Aux + 11,
+	[PA_CHANNEL_POSITION_AUX12] = ACP_CHANNEL_START_Aux + 12,
+	[PA_CHANNEL_POSITION_AUX13] = ACP_CHANNEL_START_Aux + 12,
+	[PA_CHANNEL_POSITION_AUX14] = ACP_CHANNEL_START_Aux + 14,
+	[PA_CHANNEL_POSITION_AUX15] = ACP_CHANNEL_START_Aux + 15,
+	[PA_CHANNEL_POSITION_AUX16] = ACP_CHANNEL_START_Aux + 16,
+	[PA_CHANNEL_POSITION_AUX17] = ACP_CHANNEL_START_Aux + 17,
+	[PA_CHANNEL_POSITION_AUX18] = ACP_CHANNEL_START_Aux + 18,
+	[PA_CHANNEL_POSITION_AUX19] = ACP_CHANNEL_START_Aux + 19,
+	[PA_CHANNEL_POSITION_AUX20] = ACP_CHANNEL_START_Aux + 20,
+	[PA_CHANNEL_POSITION_AUX21] = ACP_CHANNEL_START_Aux + 21,
+	[PA_CHANNEL_POSITION_AUX22] = ACP_CHANNEL_START_Aux + 22,
+	[PA_CHANNEL_POSITION_AUX23] = ACP_CHANNEL_START_Aux + 23,
+	[PA_CHANNEL_POSITION_AUX24] = ACP_CHANNEL_START_Aux + 24,
+	[PA_CHANNEL_POSITION_AUX25] = ACP_CHANNEL_START_Aux + 25,
+	[PA_CHANNEL_POSITION_AUX26] = ACP_CHANNEL_START_Aux + 26,
+	[PA_CHANNEL_POSITION_AUX27] = ACP_CHANNEL_START_Aux + 27,
+	[PA_CHANNEL_POSITION_AUX28] = ACP_CHANNEL_START_Aux + 28,
+	[PA_CHANNEL_POSITION_AUX29] = ACP_CHANNEL_START_Aux + 29,
+	[PA_CHANNEL_POSITION_AUX30] = ACP_CHANNEL_START_Aux + 30,
+	[PA_CHANNEL_POSITION_AUX31] = ACP_CHANNEL_START_Aux + 31,
 
 	[PA_CHANNEL_POSITION_TOP_CENTER] = ACP_CHANNEL_TC,
 
@@ -148,8 +150,8 @@ static inline uint32_t channel_pa2acp(pa_channel_position_t channel)
 
 char *acp_channel_str(char *buf, size_t len, enum acp_channel ch)
 {
-	if (ch >= ACP_CHANNEL_CUSTOM_START) {
-		snprintf(buf, len, "AUX%d", ch - ACP_CHANNEL_CUSTOM_START);
+	if (ch >= ACP_CHANNEL_START_Aux && ch <= ACP_CHANNEL_LAST_Aux) {
+		snprintf(buf, len, "AUX%d", ch - ACP_CHANNEL_START_Aux);
 	} else if (ch >= ACP_CHANNEL_UNKNOWN && ch <= ACP_CHANNEL_BRC) {
 		snprintf(buf, len, "%s", channel_names[ch]);
 	} else {
@@ -199,8 +201,8 @@ static void device_free(void *data)
 	pa_hashmap_free(dev->ports);
 }
 
-static void init_device(pa_card *impl, pa_alsa_profile *ap, pa_alsa_device *dev,
-		pa_alsa_direction_t direction, pa_alsa_mapping *m, uint32_t index)
+static void init_device(pa_card *impl, pa_alsa_device *dev, pa_alsa_direction_t direction,
+		pa_alsa_mapping *m, uint32_t index)
 {
 	uint32_t i;
 
@@ -214,7 +216,8 @@ static void init_device(pa_card *impl, pa_alsa_profile *ap, pa_alsa_device *dev,
 	dev->device.format.format_mask = m->sample_spec.format;
 	dev->device.format.rate_mask = m->sample_spec.rate;
 	dev->device.format.channels = m->channel_map.channels;
-	pa_cvolume_reset(&dev->real_volume, m->channel_map.channels);
+	pa_cvolume_reset(&dev->real_volume, dev->device.format.channels);
+	pa_cvolume_reset(&dev->soft_volume, dev->device.format.channels);
 	for (i = 0; i < m->channel_map.channels; i++)
 		dev->device.format.map[i]= channel_pa2acp(m->channel_map.map[i]);
 	dev->direction = direction;
@@ -231,15 +234,28 @@ static void init_device(pa_card *impl, pa_alsa_profile *ap, pa_alsa_device *dev,
 		dev->device.direction = ACP_DIRECTION_CAPTURE;
 		pa_proplist_update(dev->proplist, PA_UPDATE_REPLACE, m->input_proplist);
 	}
-	pa_proplist_sets(dev->proplist, PA_PROP_DEVICE_PROFILE_NAME, ap->name);
-	pa_proplist_sets(dev->proplist, PA_PROP_DEVICE_PROFILE_DESCRIPTION, ap->description);
+	pa_proplist_sets(dev->proplist, PA_PROP_DEVICE_PROFILE_NAME, m->name);
+	pa_proplist_sets(dev->proplist, PA_PROP_DEVICE_PROFILE_DESCRIPTION, m->description);
 	pa_proplist_setf(dev->proplist, "card.profile.device", "%u", index);
 	pa_proplist_as_dict(dev->proplist, &dev->device.props);
 
 	dev->ports = pa_hashmap_new(pa_idxset_string_hash_func,
 			pa_idxset_string_compare_func);
-	if (m->ucm_context.ucm)
+	if (m->ucm_context.ucm) {
 		dev->ucm_context = &m->ucm_context;
+		if (impl->ucm.alibpref != NULL) {
+			char **d;
+			for (d = m->device_strings; *d; d++) {
+				if (pa_startswith(*d, impl->ucm.alibpref)) {
+					size_t plen = strlen(impl->ucm.alibpref);
+					size_t len = strlen(*d);
+					memmove(*d, (*d) + plen, len - plen + 1);
+					dev->device.flags |= ACP_DEVICE_UCM_DEVICE;
+					break;
+				}
+			}
+		}
+	}
 	pa_dynarray_init(&dev->port_array, NULL);
 }
 
@@ -302,8 +318,10 @@ static int add_pro_profile(pa_card *impl, uint32_t index)
 	if ((err = snd_ctl_open(&ctl_hndl, device, 0)) < 0) {
 		pa_log_error("can't open control for card %s: %s",
 				device, snd_strerror(err));
+	        free(device);
 		return err;
 	}
+	free(device);
 
 	snd_pcm_info_alloca(&pcminfo);
 
@@ -430,7 +448,7 @@ static void add_profiles(pa_card *impl)
 			PA_IDXSET_FOREACH(m, ap->output_mappings, idx) {
 				dev = &m->output;
 				if (dev->mapping == NULL) {
-					init_device(impl, ap, dev, PA_ALSA_DIRECTION_OUTPUT, m, n_devices++);
+					init_device(impl, dev, PA_ALSA_DIRECTION_OUTPUT, m, n_devices++);
 					pa_dynarray_append(&impl->out.devices, dev);
 				}
 				if (impl->use_ucm) {
@@ -451,7 +469,7 @@ static void add_profiles(pa_card *impl)
 			PA_IDXSET_FOREACH(m, ap->input_mappings, idx) {
 				dev = &m->input;
 				if (dev->mapping == NULL) {
-					init_device(impl, ap, dev, PA_ALSA_DIRECTION_INPUT, m, n_devices++);
+					init_device(impl, dev, PA_ALSA_DIRECTION_INPUT, m, n_devices++);
 					pa_dynarray_append(&impl->out.devices, dev);
 				}
 
@@ -549,7 +567,7 @@ static pa_available_t calc_port_state(pa_device_port *p, pa_card *impl)
 				break;
 			}
 
-			/* If the current availablility is unknown go the more precise no,
+			/* If the current availability is unknown go the more precise no,
 			* but otherwise don't change state */
 			if (pa == PA_AVAILABLE_UNKNOWN)
 				pa = cpa;
@@ -866,7 +884,7 @@ static int hdmi_eld_changed(snd_mixer_elem_t *melem, unsigned int mask)
 		changed |= old_monitor_name != NULL;
 		pa_proplist_unset(p->proplist, PA_PROP_DEVICE_PRODUCT_NAME);
 	} else {
-		changed |= (old_monitor_name == NULL) || (strcmp(old_monitor_name, eld.monitor_name) != 0);
+		changed |= (old_monitor_name == NULL) || (!spa_streq(old_monitor_name, eld.monitor_name));
 		pa_proplist_sets(p->proplist, PA_PROP_DEVICE_PRODUCT_NAME, eld.monitor_name);
 	}
 	pa_proplist_as_dict(p->proplist, &p->port.props);
@@ -937,7 +955,7 @@ uint32_t acp_card_find_best_profile_index(struct acp_card *card, const char *nam
 		struct acp_card_profile *p = profiles[i];
 
 		if (name) {
-			if (strcmp(name, p->name) == 0)
+			if (spa_streq(name, p->name))
 				best = i;
 		} else if (p->flags & ACP_PROFILE_OFF) {
 			off = i;
@@ -1029,14 +1047,19 @@ static int read_volume(pa_alsa_device *dev)
 	/* Shift down by the base volume, so that 0dB becomes maximum volume */
 	pa_sw_cvolume_multiply_scalar(&r, &r, dev->base_volume);
 
-	if (pa_cvolume_equal(&dev->real_volume, &r))
+	if (pa_cvolume_equal(&dev->hardware_volume, &r))
 		return 0;
 
-	dev->real_volume = r;
-	pa_log_info("New hardware volume:");
+	dev->real_volume = dev->hardware_volume = r;
+
+	pa_log_info("New hardware volume: min:%d max:%d",
+			pa_cvolume_min(&r), pa_cvolume_max(&r));
+
 	for (i = 0; i < r.channels; i++)
 		pa_log_debug("  %d: %d", i, r.values[i]);
 
+	pa_cvolume_reset(&dev->soft_volume, r.channels);
+
 	if (impl->events && impl->events->volume_changed)
 		impl->events->volume_changed(impl->user_data, &dev->device);
 
@@ -1164,7 +1187,7 @@ static void mixer_volume_init(pa_card *impl, pa_alsa_device *dev)
 		pa_log_info("Using hardware volume control. Hardware dB scale %s.",
 				dev->mixer_path->has_dB ? "supported" : "not supported");
 	}
-	dev->device.base_volume = pa_sw_volume_to_linear(dev->base_volume);;
+	dev->device.base_volume = pa_sw_volume_to_linear(dev->base_volume);
 	dev->device.volume_step = 1.0f / dev->n_volume_steps;
 
 	if (impl->soft_mixer || !dev->mixer_path || !dev->mixer_path->has_mute) {
@@ -1313,8 +1336,15 @@ static int device_enable(pa_card *impl, pa_alsa_mapping *mapping, pa_alsa_device
 
 	if (dev->read_volume)
 		dev->read_volume(dev);
+	else {
+		pa_cvolume_reset(&dev->real_volume, dev->device.format.channels);
+		pa_cvolume_reset(&dev->soft_volume, dev->device.format.channels);
+	}
 	if (dev->read_mute)
 		dev->read_mute(dev);
+	else
+		dev->muted = false;
+
 	return 0;
 }
 
@@ -1436,7 +1466,7 @@ static const char *acp_dict_lookup(const struct acp_dict *dict, const char *key)
 {
 	const struct acp_dict_item *it;
 	acp_dict_for_each(it, dict) {
-		if (strcmp(key, it->key) == 0)
+		if (spa_streq(key, it->key))
 			return it->value;
 	}
 	return NULL;
@@ -1472,19 +1502,19 @@ struct acp_card *acp_card_new(uint32_t index, const struct acp_dict *props)
 
 	if (props) {
 		if ((s = acp_dict_lookup(props, "api.alsa.use-ucm")) != NULL)
-			impl->use_ucm = (strcmp(s, "true") == 0 || atoi(s) == 1);
+			impl->use_ucm = spa_atob(s);
 		if ((s = acp_dict_lookup(props, "api.alsa.soft-mixer")) != NULL)
-			impl->soft_mixer = (strcmp(s, "true") == 0 || atoi(s) == 1);
+			impl->soft_mixer = spa_atob(s);
 		if ((s = acp_dict_lookup(props, "api.alsa.ignore-dB")) != NULL)
-			ignore_dB = (strcmp(s, "true") == 0 || atoi(s) == 1);
+			ignore_dB = spa_atob(s);
 		if ((s = acp_dict_lookup(props, "device.profile-set")) != NULL)
 			profile_set = s;
 		if ((s = acp_dict_lookup(props, "device.profile")) != NULL)
 			profile = s;
 		if ((s = acp_dict_lookup(props, "api.acp.auto-profile")) != NULL)
-			impl->auto_profile = (strcmp(s, "true") == 0 || atoi(s) == 1);
+			impl->auto_profile = spa_atob(s);
 		if ((s = acp_dict_lookup(props, "api.acp.auto-port")) != NULL)
-			impl->auto_port = (strcmp(s, "true") == 0 || atoi(s) == 1);
+			impl->auto_port = spa_atob(s);
 	}
 
 	impl->ucm.default_sample_spec.format = PA_SAMPLE_S16NE;
@@ -1709,7 +1739,7 @@ uint32_t acp_device_find_best_port_index(struct acp_device *dev, const char *nam
 		struct acp_port *p = ports[i];
 
 		if (name) {
-			if (strcmp(name, p->name) == 0)
+			if (spa_streq(name, p->name))
 				best = i;
 		} else if (p->available == ACP_AVAILABLE_YES) {
 			if (best == ACP_INVALID_INDEX || p->priority > ports[best]->priority)
@@ -1801,9 +1831,12 @@ int acp_device_set_volume(struct acp_device *dev, const float *volume, uint32_t
 
 	v.channels = d->mapping->channel_map.channels;
 	for (i = 0; i < v.channels; i++)
-		v.values[i] = pa_sw_volume_from_linear(volume[i % n_volume]);;
+		v.values[i] = pa_sw_volume_from_linear(volume[i % n_volume]);
+
+	pa_log_info("Set %s volume: min:%d max:%d",
+			d->set_volume ? "hardware" : "software",
+			pa_cvolume_min(&v), pa_cvolume_max(&v));
 
-	pa_log_info("Set %s volume: %d", d->set_volume ? "hardware" : "software", pa_cvolume_max(&v));
 	for (i = 0; i < v.channels; i++)
 		pa_log_debug("  %d: %d", i, v.values[i]);
 
