@@ -98,7 +98,11 @@ enum acp_channel {
 	ACP_CHANNEL_BC,			/**< bottom center */
 	ACP_CHANNEL_BLC,		/**< bottom left center */
 	ACP_CHANNEL_BRC,		/**< bottom right center */
-	ACP_CHANNEL_CUSTOM_START  = 0x10000,
+
+	ACP_CHANNEL_START_Aux  = 0x1000,
+	ACP_CHANNEL_LAST_Aux  = 0x1fff,
+
+	ACP_CHANNEL_START_Custom  = 0x10000,
 };
 
 char *acp_channel_str(char *buf, size_t len, enum acp_channel ch);
@@ -177,10 +181,6 @@ struct acp_card_events {
 
 	void (*volume_changed) (void *data, struct acp_device *dev);
 	void (*mute_changed) (void *data, struct acp_device *dev);
-
-	void (*set_soft_volume) (void *data, struct acp_device *dev,
-			const float *volume, uint32_t n_volume);
-	void (*set_soft_mute) (void *data, struct acp_device *dev, bool mute);
 };
 
 struct acp_port {
@@ -208,6 +208,7 @@ struct acp_device {
 #define ACP_DEVICE_ACTIVE	(1<<0)
 #define ACP_DEVICE_HW_VOLUME	(1<<1)
 #define ACP_DEVICE_HW_MUTE	(1<<2)
+#define ACP_DEVICE_UCM_DEVICE	(1<<3)
 	uint32_t flags;
 
 	const char *name;
@@ -224,6 +225,8 @@ struct acp_device {
 
 	uint32_t n_ports;
 	struct acp_port **ports;
+
+	int64_t latency_ns;
 };
 
 struct acp_card_profile {
@@ -282,6 +285,7 @@ uint32_t acp_device_find_best_port_index(struct acp_device *dev, const char *nam
 int acp_device_set_port(struct acp_device *dev, uint32_t port_index, uint32_t flags);
 
 int acp_device_set_volume(struct acp_device *dev, const float *volume, uint32_t n_volume);
+int acp_device_get_soft_volume(struct acp_device *dev, float *volume, uint32_t n_volume);
 int acp_device_get_volume(struct acp_device *dev, float *volume, uint32_t n_volume);
 int acp_device_set_mute(struct acp_device *dev, bool mute);
 int acp_device_get_mute(struct acp_device *dev, bool *mute);
