@@ -393,13 +393,25 @@ static struct spa_pod *build_route(struct spa_pod_builder *b, uint32_t id,
 		0);
 	spa_pod_builder_prop(b, SPA_PARAM_ROUTE_info, SPA_POD_PROP_FLAG_HINT_DICT);
 	spa_pod_builder_push_struct(b, &f[1]);
-	spa_pod_builder_int(b, p->props.n_items);
+	spa_pod_builder_int(b, p->props.n_items + (dev ? 2 : 0));
 	acp_dict_for_each(item, &p->props) {
 		spa_pod_builder_add(b,
 				SPA_POD_String(item->key),
 				SPA_POD_String(item->value),
 				NULL);
 	}
+	if (dev != NULL) {
+		const char *str;
+		str = SPA_FLAG_IS_SET(dev->flags, ACP_DEVICE_HW_MUTE) ? "true" : "false";
+		spa_pod_builder_add(b,
+				SPA_POD_String("route.hw-mute"),
+				SPA_POD_String(str), NULL);
+
+		str = SPA_FLAG_IS_SET(dev->flags, ACP_DEVICE_HW_VOLUME) ? "true" : "false";
+		spa_pod_builder_add(b,
+				SPA_POD_String("route.hw-volume"),
+				SPA_POD_String(str), NULL);
+	}
 	spa_pod_builder_pop(b, &f[1]);
 	spa_pod_builder_prop(b, SPA_PARAM_ROUTE_profiles, 0);
 	spa_pod_builder_push_array(b, &f[1]);
@@ -448,6 +460,9 @@ static struct spa_pod *build_route(struct spa_pod_builder *b, uint32_t id,
 		spa_pod_builder_array(b, sizeof(float), SPA_TYPE_Float,
 				channels, soft_volumes);
 
+		spa_pod_builder_prop(b, SPA_PROP_latencyOffsetNsec, 0);
+		spa_pod_builder_long(b, dev->latency_ns);
+
 		spa_pod_builder_pop(b, &f[1]);
 	}
 	spa_pod_builder_prop(b, SPA_PARAM_ROUTE_devices, 0);
@@ -562,6 +577,32 @@ static int impl_enum_params(void *object, int seq,
 	return 0;
 }
 
+static void on_latency_changed(void *data, struct acp_device *dev)
+{
+	struct impl *this = data;
+	struct spa_event *event;
+	uint8_t buffer[4096];
+	struct spa_pod_builder b = { 0 };
+	struct spa_pod_frame f[1];
+
+	spa_log_info(this->log, "device %s latency changed", dev->name);
+	this->info.change_mask |= SPA_DEVICE_CHANGE_MASK_PARAMS;
+	this->params[IDX_Route].user++;
+
+	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+	spa_pod_builder_push_object(&b, &f[0],
+			SPA_TYPE_EVENT_Device, SPA_DEVICE_EVENT_ObjectConfig);
+	spa_pod_builder_prop(&b, SPA_EVENT_DEVICE_Object, 0);
+	spa_pod_builder_int(&b, dev->index);
+	spa_pod_builder_prop(&b, SPA_EVENT_DEVICE_Props, 0);
+	spa_pod_builder_add_object(&b,
+			SPA_TYPE_OBJECT_Props, SPA_EVENT_DEVICE_Props,
+			SPA_PROP_latencyOffsetNsec, SPA_POD_Long(dev->latency_ns));
+	event = spa_pod_builder_pop(&b, &f[0]);
+
+	spa_device_emit_event(&this->hooks, event);
+}
+
 static int apply_device_props(struct impl *this, struct acp_device *dev, struct spa_pod *props)
 {
 	float volume = 0;
@@ -602,6 +643,20 @@ static int apply_device_props(struct impl *this, struct acp_device *dev, struct
 				changed++;
 			}
 			break;
+		case SPA_PROP_latencyOffsetNsec:
+		{
+			int64_t latency_ns;
+			if (spa_pod_get_long(&prop->value, &latency_ns) == 0) {
+				if (dev->latency_ns != latency_ns) {
+					dev->latency_ns = latency_ns;
+					on_latency_changed(this, dev);
+					changed++;
+				}
+			}
+			break;
+		}
+		default:
+			break;
 		}
 	}
 	if (n_volumes > 0)
@@ -622,28 +677,28 @@ static int impl_set_param(void *object,
 	switch (id) {
 	case SPA_PARAM_Profile:
 	{
-		uint32_t id;
+		uint32_t idx;
 		bool save = false;
 
 		if (param == NULL) {
-			id = acp_card_find_best_profile_index(this->card, NULL);
+			idx = acp_card_find_best_profile_index(this->card, NULL);
 			save = true;
 		} else if ((res = spa_pod_parse_object(param,
 				SPA_TYPE_OBJECT_ParamProfile, NULL,
-				SPA_PARAM_PROFILE_index, SPA_POD_Int(&id),
+				SPA_PARAM_PROFILE_index, SPA_POD_Int(&idx),
 				SPA_PARAM_PROFILE_save, SPA_POD_OPT_Bool(&save))) < 0) {
 			spa_log_warn(this->log, "can't parse profile");
 			spa_debug_pod(0, NULL, param);
 			return res;
 		}
 
-		acp_card_set_profile(this->card, id, save ? ACP_PROFILE_SAVE : 0);
+		acp_card_set_profile(this->card, idx, save ? ACP_PROFILE_SAVE : 0);
 		emit_info(this, false);
 		break;
 	}
 	case SPA_PARAM_Route:
 	{
-		uint32_t id, device;
+		uint32_t idx, device;
 		struct spa_pod *props = NULL;
 		struct acp_device *dev;
 		bool save = false;
@@ -653,7 +708,7 @@ static int impl_set_param(void *object,
 
 		if ((res = spa_pod_parse_object(param,
 				SPA_TYPE_OBJECT_ParamRoute, NULL,
-				SPA_PARAM_ROUTE_index, SPA_POD_Int(&id),
+				SPA_PARAM_ROUTE_index, SPA_POD_Int(&idx),
 				SPA_PARAM_ROUTE_device, SPA_POD_Int(&device),
 				SPA_PARAM_ROUTE_props, SPA_POD_OPT_Pod(&props),
 				SPA_PARAM_ROUTE_save, SPA_POD_OPT_Bool(&save))) < 0) {
@@ -665,7 +720,7 @@ static int impl_set_param(void *object,
 			return -EINVAL;
 
 		dev = this->card->devices[device];
-		acp_device_set_port(dev, id, save ? ACP_PORT_SAVE : 0);
+		acp_device_set_port(dev, idx, save ? ACP_PORT_SAVE : 0);
 		if (props)
 			apply_device_props(this, dev, props);
 		emit_info(this, false);
@@ -864,7 +919,7 @@ static void on_mute_changed(void *data, struct acp_device *dev)
 	spa_device_emit_event(&this->hooks, event);
 }
 
-struct acp_card_events card_events = {
+static const struct acp_card_events card_events = {
 	ACP_VERSION_CARD_EVENTS,
 	.props_changed = card_props_changed,
 	.profile_changed = card_profile_changed,
