@@ -9,12 +9,71 @@
 #include <limits.h>
 
 #include <spa/pod/filter.h>
+#include <spa/utils/string.h>
 #include <spa/support/system.h>
 
 #define NAME "alsa-pcm"
 
 #include "alsa-pcm.h"
 
+int spa_alsa_init(struct state *state)
+{
+	int err;
+
+	snd_config_update_free_global();
+
+	if (state->stream == SND_PCM_STREAM_PLAYBACK) {
+		state->is_iec958 = spa_strstartswith(state->props.device, "iec958");
+		state->is_hdmi = spa_strstartswith(state->props.device, "hdmi");
+	}
+
+	if (state->open_ucm) {
+		char card_name[64];
+		const char *alibpref = NULL;
+
+		snprintf(card_name, sizeof(card_name), "hw:%i", state->card_index);
+		err = snd_use_case_mgr_open(&state->ucm, card_name);
+		if (err < 0) {
+			char *name;
+			err = snd_card_get_name(state->card_index, &name);
+			if (err < 0) {
+				spa_log_error(state->log,
+						"can't get card name from index %d",
+						state->card_index);
+				return err;
+			}
+			snprintf(card_name, sizeof(card_name), "%s", name);
+			free(name);
+
+			err = snd_use_case_mgr_open(&state->ucm, card_name);
+			if (err < 0) {
+				spa_log_error(state->log, "UCM not available for card %s", card_name);
+				return err;
+			}
+		}
+
+		if ((snd_use_case_get(state->ucm, "_alibpref", &alibpref) != 0))
+			alibpref = NULL;
+		if (alibpref != NULL) {
+			char name[sizeof(state->props.device)];
+			spa_scnprintf(name, sizeof(name), "%s%s", alibpref,
+					state->props.device);
+			strcpy(state->props.device, name);
+			free((void*)alibpref);
+		}
+	}
+
+	return 0;
+}
+
+int spa_alsa_clear(struct state *state)
+{
+	if (state->ucm)
+		snd_use_case_mgr_close(state->ucm);
+	state->ucm = NULL;
+	return 0;
+}
+
 #define CHECK(s,msg,...) if ((err = (s)) < 0) { spa_log_error(state->log, msg ": %s", ##__VA_ARGS__, snd_strerror(err)); return err; }
 
 int spa_alsa_open(struct state *state)
@@ -260,40 +319,21 @@ static void sanitize_map(snd_pcm_chmap_t* map)
 	}
 }
 
-int
-spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
-		     const struct spa_pod *filter)
+static int enum_pcm_formats(struct state *state, uint32_t index, uint32_t *next,
+		struct spa_pod **result, struct spa_pod_builder *b)
 {
+	int err, dir;
+	size_t i, j;
 	snd_pcm_t *hndl;
 	snd_pcm_hw_params_t *params;
+	snd_pcm_chmap_query_t **maps;
+	struct spa_pod_frame f[2];
 	snd_pcm_format_mask_t *fmask;
 	snd_pcm_access_mask_t *amask;
-	snd_pcm_chmap_query_t **maps;
-	size_t i, j;
-	int err, dir;
 	unsigned int min, max;
 	unsigned int rrate, rchannels;
-	uint8_t buffer[4096];
-	struct spa_pod_builder b = { 0 };
 	struct spa_pod_choice *choice;
-	struct spa_pod *fmt;
-	int res;
-	bool opened;
-	struct spa_pod_frame f[2];
-	struct spa_result_node_params result;
-	uint32_t count = 0, rate;
-
-	opened = state->opened;
-	if ((err = spa_alsa_open(state)) < 0)
-		return err;
-
-	result.id = SPA_PARAM_EnumFormat;
-	result.next = start;
-
-      next:
-	result.index = result.next++;
-
-	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+	uint32_t rate;
 
 	hndl = state->hndl;
 	snd_pcm_hw_params_alloca(&params);
@@ -318,8 +358,8 @@ spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
 		}
 	}
 
-	spa_pod_builder_push_object(&b, &f[0], SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat);
-	spa_pod_builder_add(&b,
+	spa_pod_builder_push_object(b, &f[0], SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat);
+	spa_pod_builder_add(b,
 			SPA_FORMAT_mediaType,    SPA_POD_Id(SPA_MEDIA_TYPE_audio),
 			SPA_FORMAT_mediaSubtype, SPA_POD_Id(SPA_MEDIA_SUBTYPE_raw),
 			0);
@@ -330,10 +370,10 @@ spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
 	snd_pcm_access_mask_alloca(&amask);
 	snd_pcm_hw_params_get_access_mask(params, amask);
 
-	spa_pod_builder_prop(&b, SPA_FORMAT_AUDIO_format, 0);
+	spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_format, 0);
 
-	spa_pod_builder_push_choice(&b, &f[1], SPA_CHOICE_None, 0);
-	choice = (struct spa_pod_choice*)spa_pod_builder_frame(&b, &f[1]);
+	spa_pod_builder_push_choice(b, &f[1], SPA_CHOICE_None, 0);
+	choice = (struct spa_pod_choice*)spa_pod_builder_frame(b, &f[1]);
 
 	for (i = 1, j = 0; i < SPA_N_ELEMENTS(format_info); i++) {
 		const struct format_info *fi = &format_info[i];
@@ -344,34 +384,42 @@ spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
 			    fi->spa_pformat != SPA_AUDIO_FORMAT_UNKNOWN &&
 			    (state->default_format == 0 || state->default_format == fi->spa_pformat)) {
 				if (j++ == 0)
-					spa_pod_builder_id(&b, fi->spa_pformat);
-				spa_pod_builder_id(&b, fi->spa_pformat);
+					spa_pod_builder_id(b, fi->spa_pformat);
+				spa_pod_builder_id(b, fi->spa_pformat);
 			}
 			if ((snd_pcm_access_mask_test(amask, SND_PCM_ACCESS_MMAP_INTERLEAVED) ||
 			    snd_pcm_access_mask_test(amask, SND_PCM_ACCESS_RW_INTERLEAVED)) &&
 			    (state->default_format == 0 || state->default_format == fi->spa_format)) {
 				if (j++ == 0)
-					spa_pod_builder_id(&b, fi->spa_format);
-				spa_pod_builder_id(&b, fi->spa_format);
+					spa_pod_builder_id(b, fi->spa_format);
+				spa_pod_builder_id(b, fi->spa_format);
 			}
 		}
 	}
 	if (j == 0) {
 		char buf[1024];
-		int i, offs;
+		int i, r, offs;
 
 		for (i = 0, offs = 0; i <= SND_PCM_FORMAT_LAST; i++) {
-			if (snd_pcm_format_mask_test(fmask, (snd_pcm_format_t)i))
-				offs += snprintf(&buf[offs], sizeof(buf) - offs,
+			if (snd_pcm_format_mask_test(fmask, (snd_pcm_format_t)i)) {
+				r = snprintf(&buf[offs], sizeof(buf) - offs,
 						"%s ", snd_pcm_format_name((snd_pcm_format_t)i));
+				if (r < 0 || r + offs >= (int)sizeof(buf))
+					return -ENOSPC;
+				offs += r;
+			}
 		}
 		spa_log_warn(state->log, "%s: unsupported card: formats:%s",
 				state->props.device, buf);
 
 		for (i = 0, offs = 0; i <= SND_PCM_ACCESS_LAST; i++) {
-			if (snd_pcm_access_mask_test(amask, (snd_pcm_access_t)i))
-				offs += snprintf(&buf[offs], sizeof(buf) - offs,
+			if (snd_pcm_access_mask_test(amask, (snd_pcm_access_t)i)) {
+				r = snprintf(&buf[offs], sizeof(buf) - offs,
 						"%s ", snd_pcm_access_name((snd_pcm_access_t)i));
+				if (r < 0 || r + offs >= (int)sizeof(buf))
+					return -ENOSPC;
+				offs += r;
+			}
 		}
 		spa_log_warn(state->log, "%s: unsupported card: access:%s",
 				state->props.device, buf);
@@ -379,7 +427,7 @@ spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
 	}
 	if (j > 1)
 		choice->body.type = SPA_CHOICE_Enum;
-	spa_pod_builder_pop(&b, &f[1]);
+	spa_pod_builder_pop(b, &f[1]);
 
 	CHECK(snd_pcm_hw_params_get_rate_min(params, &min, &dir), "get_rate_min");
 	CHECK(snd_pcm_hw_params_get_rate_max(params, &max, &dir), "get_rate_max");
@@ -391,20 +439,20 @@ spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
 			max = state->default_rate;
 	}
 
-	spa_pod_builder_prop(&b, SPA_FORMAT_AUDIO_rate, 0);
+	spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_rate, 0);
 
-	spa_pod_builder_push_choice(&b, &f[1], SPA_CHOICE_None, 0);
-	choice = (struct spa_pod_choice*)spa_pod_builder_frame(&b, &f[1]);
+	spa_pod_builder_push_choice(b, &f[1], SPA_CHOICE_None, 0);
+	choice = (struct spa_pod_choice*)spa_pod_builder_frame(b, &f[1]);
 
 	rate = state->position ? state->position->clock.rate.denom : DEFAULT_RATE;
 
-	spa_pod_builder_int(&b, SPA_CLAMP(rate, min, max));
+	spa_pod_builder_int(b, SPA_CLAMP(rate, min, max));
 	if (min != max) {
-		spa_pod_builder_int(&b, min);
-		spa_pod_builder_int(&b, max);
+		spa_pod_builder_int(b, min);
+		spa_pod_builder_int(b, max);
 		choice->body.type = SPA_CHOICE_Range;
 	}
-	spa_pod_builder_pop(&b, &f[1]);
+	spa_pod_builder_pop(b, &f[1]);
 
 	CHECK(snd_pcm_hw_params_get_channels_min(params, &min), "get_channels_min");
 	CHECK(snd_pcm_hw_params_get_channels_max(params, &max), "get_channels_max");
@@ -419,55 +467,55 @@ spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
 	min = SPA_MIN(min, SPA_AUDIO_MAX_CHANNELS);
 	max = SPA_MIN(max, SPA_AUDIO_MAX_CHANNELS);
 
-	spa_pod_builder_prop(&b, SPA_FORMAT_AUDIO_channels, 0);
+	spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_channels, 0);
 
 	if (state->props.use_chmap && (maps = snd_pcm_query_chmaps(hndl)) != NULL) {
 		uint32_t channel;
 		snd_pcm_chmap_t* map;
 
 skip_channels:
-		if (maps[result.index] == NULL) {
+		if (maps[index] == NULL) {
 			snd_pcm_free_chmaps(maps);
-			goto enum_end;
+			return 0;
 		}
-		map = &maps[result.index]->map;
+		map = &maps[index]->map;
 
 		spa_log_debug(state->log, "map %d channels (%d %d)", map->channels, min, max);
 
 		if (map->channels < min || map->channels > max) {
-			result.index = result.next++;
+			index = (*next)++;
 			goto skip_channels;
 		}
 
 		sanitize_map(map);
-		spa_pod_builder_int(&b, map->channels);
+		spa_pod_builder_int(b, map->channels);
 
-		spa_pod_builder_prop(&b, SPA_FORMAT_AUDIO_position, 0);
-		spa_pod_builder_push_array(&b, &f[1]);
+		spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_position, 0);
+		spa_pod_builder_push_array(b, &f[1]);
 		for (j = 0; j < map->channels; j++) {
 			spa_log_debug(state->log, NAME" %p: position %zd %d", state, j, map->pos[j]);
 			channel = chmap_position_to_channel(map->pos[j]);
-			spa_pod_builder_id(&b, channel);
+			spa_pod_builder_id(b, channel);
 		}
-		spa_pod_builder_pop(&b, &f[1]);
+		spa_pod_builder_pop(b, &f[1]);
 
 		snd_pcm_free_chmaps(maps);
 	}
 	else {
 		const struct channel_map *map = NULL;
 
-		if (result.index > 0)
-			goto enum_end;
+		if (index > 0)
+			return 0;
 
-		spa_pod_builder_push_choice(&b, &f[1], SPA_CHOICE_None, 0);
-		choice = (struct spa_pod_choice*)spa_pod_builder_frame(&b, &f[1]);
-		spa_pod_builder_int(&b, max);
+		spa_pod_builder_push_choice(b, &f[1], SPA_CHOICE_None, 0);
+		choice = (struct spa_pod_choice*)spa_pod_builder_frame(b, &f[1]);
+		spa_pod_builder_int(b, max);
 		if (min != max) {
-			spa_pod_builder_int(&b, min);
-			spa_pod_builder_int(&b, max);
+			spa_pod_builder_int(b, min);
+			spa_pod_builder_int(b, max);
 			choice->body.type = SPA_CHOICE_Range;
 		}
-		spa_pod_builder_pop(&b, &f[1]);
+		spa_pod_builder_pop(b, &f[1]);
 
 		if (min == max) {
 			if (state->default_pos.channels == min)
@@ -476,19 +524,160 @@ skip_channels:
 				map = &default_map[min];
 		}
 		if (map) {
-			spa_pod_builder_prop(&b, SPA_FORMAT_AUDIO_position, 0);
-			spa_pod_builder_push_array(&b, &f[1]);
+			spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_position, 0);
+			spa_pod_builder_push_array(b, &f[1]);
 			for (j = 0; j < map->channels; j++) {
 				spa_log_debug(state->log, NAME" %p: position %zd %d", state, j, map->pos[j]);
-				spa_pod_builder_id(&b, map->pos[j]);
+				spa_pod_builder_id(b, map->pos[j]);
 			}
-			spa_pod_builder_pop(&b, &f[1]);
+			spa_pod_builder_pop(b, &f[1]);
 		}
 	}
+	*result = spa_pod_builder_pop(b, &f[0]);
+	return 1;
+}
 
-	fmt = spa_pod_builder_pop(&b, &f[0]);
+static bool codec_supported(uint32_t codec, unsigned int chmax, unsigned int rmax)
+{
+	switch (codec) {
+	case SPA_AUDIO_IEC958_CODEC_PCM:
+	case SPA_AUDIO_IEC958_CODEC_DTS:
+	case SPA_AUDIO_IEC958_CODEC_AC3:
+	case SPA_AUDIO_IEC958_CODEC_MPEG:
+	case SPA_AUDIO_IEC958_CODEC_MPEG2_AAC:
+		if (chmax >= 2)
+			return true;
+		break;
+	case SPA_AUDIO_IEC958_CODEC_EAC3:
+		if (rmax >= 48000 * 4 && chmax >= 2)
+			return true;
+		break;
+	case SPA_AUDIO_IEC958_CODEC_TRUEHD:
+	case SPA_AUDIO_IEC958_CODEC_DTSHD:
+		if (chmax >= 8)
+			return true;
+		break;
+	}
+	return false;
+}
 
-	if ((res = spa_pod_filter(&b, &result.param, fmt, filter)) < 0)
+static int enum_iec958_formats(struct state *state, uint32_t index, uint32_t *next,
+		struct spa_pod **result, struct spa_pod_builder *b)
+{
+	int err, dir;
+	snd_pcm_t *hndl;
+	snd_pcm_hw_params_t *params;
+	struct spa_pod_frame f[2];
+	struct spa_pod_choice *choice;
+	unsigned int rmin, rmax;
+	unsigned int chmin, chmax;
+	uint32_t i, c, codecs[16], n_codecs;
+
+	if ((index & 0xffff) > 0)
+		return 0;
+
+	if (!(state->is_iec958 || state->is_hdmi))
+		return 0;
+	if (state->iec958_codecs == 0)
+		return 0;
+
+	hndl = state->hndl;
+	snd_pcm_hw_params_alloca(&params);
+	CHECK(snd_pcm_hw_params_any(hndl, params), "Broken configuration: no configurations available");
+
+	CHECK(snd_pcm_hw_params_set_rate_resample(hndl, params, 0), "set_rate_resample");
+
+	spa_pod_builder_push_object(b, &f[0], SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat);
+	spa_pod_builder_add(b,
+			SPA_FORMAT_mediaType,    SPA_POD_Id(SPA_MEDIA_TYPE_audio),
+			SPA_FORMAT_mediaSubtype, SPA_POD_Id(SPA_MEDIA_SUBTYPE_iec958),
+			0);
+
+	CHECK(snd_pcm_hw_params_get_channels_min(params, &chmin), "get_channels_min");
+	CHECK(snd_pcm_hw_params_get_channels_max(params, &chmax), "get_channels_max");
+	spa_log_debug(state->log, "channels (%d %d)", chmin, chmax);
+
+	CHECK(snd_pcm_hw_params_get_rate_min(params, &rmin, &dir), "get_rate_min");
+	CHECK(snd_pcm_hw_params_get_rate_max(params, &rmax, &dir), "get_rate_max");
+	spa_log_debug(state->log, "rate (%d %d)", rmin, rmax);
+
+	if (state->default_rate != 0) {
+		if (rmin < state->default_rate)
+			rmin = state->default_rate;
+		if (rmax > state->default_rate)
+			rmax = state->default_rate;
+	}
+
+	spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_iec958Codec, 0);
+	spa_pod_builder_push_choice(b, &f[1], SPA_CHOICE_Enum, 0);
+
+	n_codecs = spa_alsa_get_iec958_codecs(state, codecs, SPA_N_ELEMENTS(codecs));
+	for (i = 0, c = 0; i < n_codecs; i++) {
+		if (!codec_supported(codecs[i], chmax, rmax))
+			continue;
+		if (c++ == 0)
+			spa_pod_builder_id(b, codecs[i]);
+		spa_pod_builder_id(b, codecs[i]);
+	}
+	spa_pod_builder_pop(b, &f[1]);
+
+	spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_rate, 0);
+	spa_pod_builder_push_choice(b, &f[1], SPA_CHOICE_None, 0);
+	choice = (struct spa_pod_choice*)spa_pod_builder_frame(b, &f[1]);
+
+	spa_pod_builder_int(b, SPA_CLAMP(DEFAULT_RATE, rmin, rmax));
+	if (rmin != rmax) {
+		spa_pod_builder_int(b, rmin);
+		spa_pod_builder_int(b, rmax);
+		choice->body.type = SPA_CHOICE_Range;
+	}
+	spa_pod_builder_pop(b, &f[1]);
+
+	(*next)++;
+	*result = spa_pod_builder_pop(b, &f[0]);
+	return 1;
+}
+
+int
+spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
+		     const struct spa_pod *filter)
+{
+	uint8_t buffer[4096];
+	struct spa_pod_builder b = { 0 };
+	struct spa_pod *fmt;
+	int err, res;
+	bool opened;
+	struct spa_result_node_params result;
+	uint32_t count = 0;
+
+	opened = state->opened;
+	if ((err = spa_alsa_open(state)) < 0)
+		return err;
+
+	result.id = SPA_PARAM_EnumFormat;
+	result.next = start;
+
+      next:
+	result.index = result.next++;
+
+	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+
+	if (result.index < 0x10000) {
+		if ((res = enum_pcm_formats(state, result.index, &result.next, &fmt, &b)) != 1) {
+			result.next = 0x10000;
+			goto next;
+		}
+	}
+	else if (result.index < 0x20000) {
+		if ((res = enum_iec958_formats(state, result.index, &result.next, &fmt, &b)) != 1) {
+			result.next = 0x20000;
+			goto next;
+		}
+	}
+	else
+		goto enum_end;
+
+	if (spa_pod_filter(&b, &result.param, fmt, filter) < 0)
 		goto next;
 
 	spa_node_emit_result(&state->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
@@ -503,7 +692,7 @@ skip_channels:
 	return res;
 }
 
-int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_t flags)
+static int set_pcm_format(struct state *state, struct spa_audio_info_raw *info, uint32_t flags)
 {
 	unsigned int rrate, rchannels;
 	snd_pcm_uframes_t period_size;
@@ -511,7 +700,6 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 	snd_pcm_hw_params_t *params;
 	snd_pcm_format_t format;
 	snd_pcm_access_mask_t *amask;
-	struct spa_audio_info_raw *info = &fmt->info.raw;
 	snd_pcm_t *hndl;
 	unsigned int periods;
 	bool match = true, planar, is_batch;
@@ -627,19 +815,25 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 	if (is_batch)
 		state->headroom += period_size;
 
+	state->headroom = SPA_MIN(state->headroom, state->buffer_frames);
+	state->start_delay = state->default_start_delay;
+
+	state->latency[state->port_direction].min_rate = state->headroom;
+	state->latency[state->port_direction].max_rate = state->headroom;
+
 	state->period_frames = period_size;
 	periods = state->buffer_frames / state->period_frames;
 
 	spa_log_info(state->log, NAME" %s (%s): format:%s access:%s-%s rate:%d channels:%d "
 			"buffer frames %lu, period frames %lu, periods %u, frame_size %zd "
-			"headroom %u",
+			"headroom %u start-delay:%u",
 			state->props.device,
 			state->stream == SND_PCM_STREAM_CAPTURE ? "capture" : "playback",
 			snd_pcm_format_name(state->format),
 			state->use_mmap ? "mmap" : "rw",
 			planar ? "planar" : "interleaved",
 			state->rate, state->channels, state->buffer_frames, state->period_frames,
-			periods, state->frame_size, state->headroom);
+			periods, state->frame_size, state->headroom, state->start_delay);
 
 	/* write the parameters to device */
 	CHECK(snd_pcm_hw_params(hndl, params), "set_hw_params");
@@ -647,6 +841,51 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 	return match ? 0 : 1;
 }
 
+static int set_iec958_format(struct state *state, struct spa_audio_info_iec958 *info, uint32_t flags)
+{
+	struct spa_audio_info_raw fmt;
+
+	fmt.format = SPA_AUDIO_FORMAT_S16_LE;
+	fmt.channels = 2;
+	fmt.rate = info->rate;
+
+	switch (info->codec) {
+	case SPA_AUDIO_IEC958_CODEC_PCM:
+	case SPA_AUDIO_IEC958_CODEC_DTS:
+	case SPA_AUDIO_IEC958_CODEC_AC3:
+	case SPA_AUDIO_IEC958_CODEC_MPEG:
+	case SPA_AUDIO_IEC958_CODEC_MPEG2_AAC:
+		break;
+	case SPA_AUDIO_IEC958_CODEC_EAC3:
+		fmt.rate *= 4;
+		break;
+	case SPA_AUDIO_IEC958_CODEC_TRUEHD:
+	case SPA_AUDIO_IEC958_CODEC_DTSHD:
+		fmt.channels = 8;
+		break;
+	default:
+		return -ENOTSUP;
+	}
+	return set_pcm_format(state, &fmt, flags);
+}
+
+int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_t flags)
+{
+	int res;
+
+	switch (fmt->media_subtype) {
+	case SPA_MEDIA_SUBTYPE_raw:
+		res = set_pcm_format(state, &fmt->info.raw, flags);
+		break;
+	case SPA_MEDIA_SUBTYPE_iec958:
+		res = set_iec958_format(state, &fmt->info.iec958, flags);
+		break;
+	default:
+		return -ENOTSUP;
+	}
+	return res;
+}
+
 static int set_swparams(struct state *state)
 {
 	snd_pcm_t *hndl = state->hndl;
@@ -805,7 +1044,7 @@ recover:
 	state->alsa_started = false;
 
 	if (state->stream == SND_PCM_STREAM_PLAYBACK)
-		spa_alsa_silence(state, state->last_threshold * 2 + state->headroom);
+		spa_alsa_silence(state, state->start_delay + state->threshold * 2 + state->headroom);
 
 	return do_start(state);
 }
@@ -827,7 +1066,6 @@ static int get_status(struct state *state, snd_pcm_uframes_t *delay, snd_pcm_ufr
 		state->alsa_recovering = false;
 	}
 
-
 	*target = state->threshold + state->headroom;
 
 	if (state->resample && state->rate_match) {
@@ -914,6 +1152,43 @@ static int update_time(struct state *state, uint64_t nsec, snd_pcm_sframes_t del
 	return 0;
 }
 
+static inline bool is_following(struct state *state)
+{
+	return state->position && state->clock && state->position->clock.id != state->clock->id;
+}
+
+static int setup_matching(struct state *state)
+{
+	int card;
+
+	state->matching = state->following;
+
+	if (state->position == NULL)
+		return -ENOTSUP;
+
+	spa_log_debug(state->log, "clock:%s card:%d", state->position->clock.name, state->card);
+	if (sscanf(state->position->clock.name, "api.alsa.%d", &card) == 1 &&
+	    card == state->card) {
+		state->matching = false;
+	}
+	state->resample = ((uint32_t)state->rate != state->rate_denom) || state->matching;
+	return 0;
+}
+
+static inline void check_position_config(struct state *state)
+{
+	if (SPA_UNLIKELY(state->position  == NULL))
+		return;
+
+	if (SPA_UNLIKELY((state->duration != state->position->clock.duration) ||
+	    (state->rate_denom != state->position->clock.rate.denom))) {
+		state->duration = state->position->clock.duration;
+		state->rate_denom = state->position->clock.rate.denom;
+		state->threshold = (state->duration * state->rate + state->rate_denom-1) / state->rate_denom;
+		state->resample = ((uint32_t)state->rate != state->rate_denom) || state->matching;
+	}
+}
+
 int spa_alsa_write(struct state *state)
 {
 	snd_pcm_t *hndl = state->hndl;
@@ -922,10 +1197,7 @@ int spa_alsa_write(struct state *state)
 	snd_pcm_sframes_t commitres;
 	int res = 0;
 
-	if (SPA_LIKELY(state->position && state->duration != state->position->clock.duration)) {
-		state->duration = state->position->clock.duration;
-		state->threshold = (state->duration * state->rate + state->rate_denom-1) / state->rate_denom;
-	}
+	check_position_config(state);
 
 	if (state->following && state->alsa_started) {
 		uint64_t nsec;
@@ -1002,7 +1274,7 @@ again:
 
 		if (SPA_LIKELY(state->use_mmap)) {
 			for (i = 0; i < b->buf->n_datas; i++) {
-				dst = SPA_MEMBER(my_areas[i].addr, off * state->frame_size, uint8_t);
+				dst = SPA_PTROFF(my_areas[i].addr, off * state->frame_size, uint8_t);
 				src = d[i].data;
 
 				spa_memcpy(dst, src + offs, l0);
@@ -1014,7 +1286,7 @@ again:
 				void *bufs[b->buf->n_datas];
 
 				for (i = 0; i < b->buf->n_datas; i++)
-					bufs[i] = SPA_MEMBER(d[i].data, offs, void);
+					bufs[i] = SPA_PTROFF(d[i].data, offs, void);
 				snd_pcm_writen(hndl, bufs, l0 / state->frame_size);
 				if (SPA_UNLIKELY(l1 > 0)) {
 					for (i = 0; i < b->buf->n_datas; i++)
@@ -1125,10 +1397,10 @@ push_frames(struct state *state,
 			l1 = n_bytes - l0;
 
 			for (i = 0; i < b->buf->n_datas; i++) {
-				src = SPA_MEMBER(my_areas[i].addr, offset * state->frame_size, uint8_t);
+				src = SPA_PTROFF(my_areas[i].addr, offset * state->frame_size, uint8_t);
 				spa_memcpy(d[i].data, src, l0);
 				if (l1 > 0)
-					spa_memcpy(SPA_MEMBER(d[i].data, l0, void), my_areas[i].addr, l1);
+					spa_memcpy(SPA_PTROFF(d[i].data, l0, void), my_areas[i].addr, l1);
 				d[i].chunk->offset = 0;
 				d[i].chunk->size = n_bytes;
 				d[i].chunk->stride = state->frame_size;
@@ -1147,13 +1419,16 @@ push_frames(struct state *state,
 				snd_pcm_readi(state->hndl, bufs[0], total_frames);
 			}
 		}
+		spa_log_trace_fp(state->log, NAME" %p: wrote %ld frames into buffer %d",
+				state, total_frames, b->id);
+
 		spa_list_append(&state->ready, &b->link);
 	}
 	return total_frames;
 }
 
 
-int spa_alsa_read(struct state *state, snd_pcm_uframes_t silence)
+int spa_alsa_read(struct state *state)
 {
 	snd_pcm_t *hndl = state->hndl;
 	snd_pcm_uframes_t total_read = 0, to_read;
@@ -1163,10 +1438,8 @@ int spa_alsa_read(struct state *state, snd_pcm_uframes_t silence)
 	int res = 0;
 
 	if (state->position) {
-		if (state->duration != state->position->clock.duration) {
-			state->duration = state->position->clock.duration;
-			state->threshold = (state->duration * state->rate + state->rate_denom-1) / state->rate_denom;
-		}
+		check_position_config(state);
+
 		if (!state->following) {
 			uint64_t position;
 
@@ -1255,6 +1528,40 @@ int spa_alsa_read(struct state *state, snd_pcm_uframes_t silence)
 	return 0;
 }
 
+int spa_alsa_skip(struct state *state)
+{
+	struct buffer *b;
+	struct spa_data *d;
+	uint32_t i, avail, total_frames, n_bytes, frames;
+
+	if (spa_list_is_empty(&state->free)) {
+		spa_log_warn(state->log, NAME" %s: no more buffers", state->props.device);
+		return -EPIPE;
+	}
+
+	frames = state->read_size;
+
+	b = spa_list_first(&state->free, struct buffer, link);
+	spa_list_remove(&b->link);
+
+	d = b->buf->datas;
+
+	avail = d[0].maxsize / state->frame_size;
+	total_frames = SPA_MIN(avail, frames);
+	n_bytes = total_frames * state->frame_size;
+
+	for (i = 0; i < b->buf->n_datas; i++) {
+		memset(d[i].data, 0, n_bytes);
+		d[i].chunk->offset = 0;
+		d[i].chunk->size = n_bytes;
+		d[i].chunk->stride = state->frame_size;
+	}
+	spa_list_append(&state->ready, &b->link);
+
+	return 0;
+}
+
+
 static int handle_play(struct state *state, uint64_t nsec,
 		snd_pcm_uframes_t delay, snd_pcm_uframes_t target)
 {
@@ -1300,14 +1607,15 @@ static int handle_capture(struct state *state, uint64_t nsec,
 	if (SPA_UNLIKELY(res = update_time(state, nsec, delay, target, false)) < 0)
 		return res;
 
-	if ((res = spa_alsa_read(state, target)) < 0)
+	if ((res = spa_alsa_read(state)) < 0)
 		return res;
 
 	if (spa_list_is_empty(&state->ready))
 		return 0;
 
 	io = state->io;
-	if (io != NULL && io->status != SPA_STATUS_HAVE_DATA) {
+	if (io != NULL &&
+	    (io->status != SPA_STATUS_HAVE_DATA || state->rate_match != NULL)) {
 		struct buffer *b;
 
 		if (io->buffer_id < state->n_buffers)
@@ -1330,17 +1638,13 @@ static void alsa_on_timeout_event(struct spa_source *source)
 	struct state *state = source->data;
 	snd_pcm_uframes_t delay, target;
 	uint64_t expire;
-	int res;
 
 	if (SPA_UNLIKELY(state->started && spa_system_timerfd_read(state->data_system, state->timerfd, &expire) < 0))
 		spa_log_warn(state->log, NAME" %p: error reading timerfd: %m", state);
 
-	if (SPA_UNLIKELY(state->position && state->duration != state->position->clock.duration)) {
-		state->duration = state->position->clock.duration;
-		state->threshold = (state->duration * state->rate + state->rate_denom-1) / state->rate_denom;
-	}
+	check_position_config(state);
 
-	if (SPA_UNLIKELY((res = get_status(state, &delay, &target)) < 0))
+	if (SPA_UNLIKELY(get_status(state, &delay, &target) < 0))
 		return;
 
 	state->current_time = state->next_time;
@@ -1349,7 +1653,8 @@ static void alsa_on_timeout_event(struct spa_source *source)
 	if (SPA_UNLIKELY(spa_log_level_enabled(state->log, SPA_LOG_LEVEL_TRACE))) {
 		struct timespec now;
 		uint64_t nsec;
-		spa_system_clock_gettime(state->data_system, CLOCK_MONOTONIC, &now);
+		if (spa_system_clock_gettime(state->data_system, CLOCK_MONOTONIC, &now) < 0)
+		    return;
 		nsec = SPA_TIMESPEC_TO_NSEC(&now);
 		spa_log_trace_fp(state->log, NAME" %p: timeout %lu %lu %"PRIu64" %"PRIu64" %"PRIi64
 				" %d %"PRIi64, state, delay, target, nsec, state->current_time,
@@ -1387,7 +1692,10 @@ static void reset_buffers(struct state *this)
 static int set_timers(struct state *state)
 {
 	struct timespec now;
-	spa_system_clock_gettime(state->data_system, CLOCK_MONOTONIC, &now);
+	int res;
+
+	if ((res = spa_system_clock_gettime(state->data_system, CLOCK_MONOTONIC, &now)) < 0)
+	    return res;
 	state->next_time = SPA_TIMESPEC_TO_NSEC(&now);
 
 	if (state->following) {
@@ -1398,29 +1706,6 @@ static int set_timers(struct state *state)
 	return 0;
 }
 
-static inline bool is_following(struct state *state)
-{
-	return state->position && state->clock && state->position->clock.id != state->clock->id;
-}
-
-static int setup_matching(struct state *state)
-{
-	int card;
-
-	state->matching = state->following;
-
-	if (state->position == NULL)
-		return -ENOTSUP;
-
-	spa_log_debug(state->log, "clock:%s card:%d", state->position->clock.name, state->card);
-	if (sscanf(state->position->clock.name, "api.alsa.%d", &card) == 1 &&
-	    card == state->card) {
-		state->matching = false;
-	}
-	state->resample = (state->rate != state->rate_denom) || state->matching;
-	return 0;
-}
-
 int spa_alsa_start(struct state *state)
 {
 	int err;
@@ -1475,7 +1760,7 @@ int spa_alsa_start(struct state *state)
 	state->alsa_started = false;
 
 	if (state->stream == SND_PCM_STREAM_PLAYBACK)
-		spa_alsa_silence(state, state->threshold * 2 + state->headroom);
+		spa_alsa_silence(state, state->start_delay + state->threshold * 2 + state->headroom);
 
 	if ((err = do_start(state)) < 0)
 		return err;
@@ -1502,7 +1787,7 @@ static int do_reassign_follower(struct spa_loop *loop,
 
 int spa_alsa_reassign_follower(struct state *state)
 {
-	bool following;
+	bool following, freewheel;
 
 	if (!state->started)
 		return 0;
@@ -1514,6 +1799,18 @@ int spa_alsa_reassign_follower(struct state *state)
 		spa_loop_invoke(state->data_loop, do_reassign_follower, 0, NULL, 0, true, state);
 	}
 	setup_matching(state);
+
+	freewheel = state->position &&
+		SPA_FLAG_IS_SET(state->position->clock.flags, SPA_IO_CLOCK_FLAG_FREEWHEEL);
+
+	if (state->freewheel != freewheel) {
+		spa_log_debug(state->log, NAME" %p: freewheel %d->%d", state, state->freewheel, freewheel);
+		state->freewheel = freewheel;
+		if (freewheel)
+			snd_pcm_pause(state->hndl, 1);
+		else
+			snd_pcm_pause(state->hndl, 0);
+	}
 	return 0;
 }
 
