@@ -23,6 +23,7 @@
  */
 
 #include <math.h>
+#include <byteswap.h>
 
 #include <spa/utils/defs.h>
 
@@ -33,12 +34,21 @@
 #define U8_TO_F32(v)	((((uint8_t)(v)) * (1.0f / U8_OFFS)) - 1.0)
 #define F32_TO_U8(v)	(uint8_t)((SPA_CLAMP(v, -1.0f, 1.0f) * U8_SCALE) + U8_OFFS)
 
+#define S8_MIN		-127
+#define S8_MAX		127
+#define S8_MAX_F	127.0f
+#define S8_SCALE	127.0f
+#define S8_TO_F32(v)	(((int8_t)(v)) * (1.0f / S8_SCALE))
+#define F32_TO_S8(v)	(int8_t)(SPA_CLAMP(v, -1.0f, 1.0f) * S8_SCALE)
+
 #define S16_MIN		-32767
 #define S16_MAX		32767
 #define S16_MAX_F	32767.0f
 #define S16_SCALE	32767.0f
 #define S16_TO_F32(v)	(((int16_t)(v)) * (1.0f / S16_SCALE))
+#define S16S_TO_F32(v)	(((int16_t)bswap_16((uint16_t)v)) * (1.0f / S16_SCALE))
 #define F32_TO_S16(v)	(int16_t)(SPA_CLAMP(v, -1.0f, 1.0f) * S16_SCALE)
+#define F32_TO_S16S(v)	((int16_t)bswap_16((uint16_t)(SPA_CLAMP(v, -1.0f, 1.0f) * S16_SCALE)))
 
 #define S24_MIN		-8388607
 #define S24_MAX		8388607
@@ -50,8 +60,15 @@
 #define S32_SCALE	2147483648.0f
 #define S32_MIN		2147483520.0f
 
-#define S32_TO_F32(v)	S24_TO_F32((v) >> 8)
+#define S32_TO_F32(v)	S24_TO_F32(((int32_t)(v)) >> 8)
+#define S32S_TO_F32(v)	S24_TO_F32(((int32_t)bswap_32(v)) >> 8)
 #define F32_TO_S32(v)	(F32_TO_S24(v) << 8)
+#define F32_TO_S32S(v)	bswap_32((F32_TO_S24(v) << 8))
+
+#define S24_32_TO_F32(v)	S32_TO_F32((v)<<8)
+#define S24_32S_TO_F32(v)	S32_TO_F32(((int32_t)bswap_32(v))<<8)
+#define F32_TO_S24_32(v)	F32_TO_S24(v)
+#define F32_TO_S24_32S(v)	bswap_32(F32_TO_S24(v))
 
 static inline int32_t read_s24(const void *src)
 {
@@ -140,13 +157,21 @@ DEFINE_FUNCTION(u8d_to_f32d, c);
 DEFINE_FUNCTION(u8_to_f32, c);
 DEFINE_FUNCTION(u8_to_f32d, c);
 DEFINE_FUNCTION(u8d_to_f32, c);
+DEFINE_FUNCTION(s8d_to_f32d, c);
+DEFINE_FUNCTION(s8_to_f32, c);
+DEFINE_FUNCTION(s8_to_f32d, c);
+DEFINE_FUNCTION(s8d_to_f32, c);
+DEFINE_FUNCTION(ulaw_to_f32d, c);
+DEFINE_FUNCTION(alaw_to_f32d, c);
 DEFINE_FUNCTION(s16d_to_f32d, c);
 DEFINE_FUNCTION(s16_to_f32, c);
 DEFINE_FUNCTION(s16_to_f32d, c);
+DEFINE_FUNCTION(s16s_to_f32d, c);
 DEFINE_FUNCTION(s16d_to_f32, c);
 DEFINE_FUNCTION(s32d_to_f32d, c);
 DEFINE_FUNCTION(s32_to_f32, c);
 DEFINE_FUNCTION(s32_to_f32d, c);
+DEFINE_FUNCTION(s32s_to_f32d, c);
 DEFINE_FUNCTION(s32d_to_f32, c);
 DEFINE_FUNCTION(s24d_to_f32d, c);
 DEFINE_FUNCTION(s24_to_f32, c);
@@ -156,19 +181,28 @@ DEFINE_FUNCTION(s24d_to_f32, c);
 DEFINE_FUNCTION(s24_32d_to_f32d, c);
 DEFINE_FUNCTION(s24_32_to_f32, c);
 DEFINE_FUNCTION(s24_32_to_f32d, c);
+DEFINE_FUNCTION(s24_32s_to_f32d, c);
 DEFINE_FUNCTION(s24_32d_to_f32, c);
 DEFINE_FUNCTION(f32d_to_u8d, c);
 DEFINE_FUNCTION(f32_to_u8, c);
 DEFINE_FUNCTION(f32_to_u8d, c);
 DEFINE_FUNCTION(f32d_to_u8, c);
+DEFINE_FUNCTION(f32d_to_s8d, c);
+DEFINE_FUNCTION(f32_to_s8, c);
+DEFINE_FUNCTION(f32_to_s8d, c);
+DEFINE_FUNCTION(f32d_to_s8, c);
+DEFINE_FUNCTION(f32d_to_alaw, c);
+DEFINE_FUNCTION(f32d_to_ulaw, c);
 DEFINE_FUNCTION(f32d_to_s16d, c);
 DEFINE_FUNCTION(f32_to_s16, c);
 DEFINE_FUNCTION(f32_to_s16d, c);
 DEFINE_FUNCTION(f32d_to_s16, c);
+DEFINE_FUNCTION(f32d_to_s16s, c);
 DEFINE_FUNCTION(f32d_to_s32d, c);
 DEFINE_FUNCTION(f32_to_s32, c);
 DEFINE_FUNCTION(f32_to_s32d, c);
 DEFINE_FUNCTION(f32d_to_s32, c);
+DEFINE_FUNCTION(f32d_to_s32s, c);
 DEFINE_FUNCTION(f32d_to_s24d, c);
 DEFINE_FUNCTION(f32_to_s24, c);
 DEFINE_FUNCTION(f32_to_s24d, c);
@@ -178,14 +212,17 @@ DEFINE_FUNCTION(f32d_to_s24_32d, c);
 DEFINE_FUNCTION(f32_to_s24_32, c);
 DEFINE_FUNCTION(f32_to_s24_32d, c);
 DEFINE_FUNCTION(f32d_to_s24_32, c);
+DEFINE_FUNCTION(f32d_to_s24_32s, c);
 DEFINE_FUNCTION(deinterleave_8, c);
 DEFINE_FUNCTION(deinterleave_16, c);
 DEFINE_FUNCTION(deinterleave_24, c);
 DEFINE_FUNCTION(deinterleave_32, c);
+DEFINE_FUNCTION(deinterleave_32s, c);
 DEFINE_FUNCTION(interleave_8, c);
 DEFINE_FUNCTION(interleave_16, c);
 DEFINE_FUNCTION(interleave_24, c);
 DEFINE_FUNCTION(interleave_32, c);
+DEFINE_FUNCTION(interleave_32s, c);
 
 #if defined(HAVE_NEON)
 DEFINE_FUNCTION(s16_to_f32d, neon);
@@ -218,3 +255,5 @@ DEFINE_FUNCTION(f32d_to_s16_4, avx2);
 DEFINE_FUNCTION(f32d_to_s16_2, avx2);
 DEFINE_FUNCTION(f32d_to_s16, avx2);
 #endif
+
+#undef DEFINE_FUNCTION
