@@ -4,7 +4,7 @@
  *
  *  Copyright (C) 2006-2010  Nokia Corporation
  *  Copyright (C) 2004-2010  Marcel Holtmann <marcel@holtmann.org>
- *
+ *  Copyright (C) 2018       Pali Rohár <pali.rohar@gmail.com>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -29,6 +29,7 @@
 #include <stddef.h>
 
 #include <spa/param/audio/format.h>
+#include <spa/param/bluetooth/audio.h>
 #include <spa/pod/pod.h>
 #include <spa/pod/builder.h>
 
@@ -168,6 +169,19 @@
 #define APTX_HD_SAMPLING_FREQ_44100     0x2
 #define APTX_HD_SAMPLING_FREQ_48000     0x1
 
+#define APTX_LL_VENDOR_ID		0x0000000a
+#define APTX_LL_VENDOR_ID2		0x000000d7
+#define APTX_LL_CODEC_ID		0x0002
+
+/**
+ * Default parameters for aptX LL (Sprint) encoder
+ */
+#define APTX_LL_TARGET_CODEC_LEVEL      180  /* target codec buffer level */
+#define APTX_LL_INITIAL_CODEC_LEVEL     360  /* initial codec buffer level */
+#define APTX_LL_SRA_MAX_RATE            50   /* x/10000 = 0.005 SRA rate */
+#define APTX_LL_SRA_AVG_TIME            1    /* SRA averaging time = 1s */
+#define APTX_LL_GOOD_WORKING_LEVEL      180  /* good working buffer level */
+
 #define LDAC_VENDOR_ID			0x0000012d
 #define LDAC_CODEC_ID			0x00aa
 
@@ -182,6 +196,17 @@
 #define LDAC_SAMPLING_FREQ_176400	0x02
 #define LDAC_SAMPLING_FREQ_192000	0x01
 
+#define FASTSTREAM_VENDOR_ID            0x0000000a
+#define FASTSTREAM_CODEC_ID             0x0001
+
+#define FASTSTREAM_DIRECTION_SINK       0x1
+#define FASTSTREAM_DIRECTION_SOURCE     0x2
+
+#define FASTSTREAM_SINK_SAMPLING_FREQ_44100     0x2
+#define FASTSTREAM_SINK_SAMPLING_FREQ_48000     0x1
+
+#define FASTSTREAM_SOURCE_SAMPLING_FREQ_16000   0x2
+
 typedef struct {
 	uint32_t vendor_id;
 	uint16_t codec_id;
@@ -240,6 +265,20 @@ typedef struct {
 	uint32_t rfa;
 } __attribute__ ((packed)) a2dp_aptx_hd_t;
 
+typedef struct {
+        a2dp_aptx_t aptx;
+        uint8_t bidirect_link:1;
+        uint8_t has_new_caps:1;
+        uint8_t reserved:6;
+} __attribute__ ((packed)) a2dp_aptx_ll_t;
+
+typedef struct {
+        a2dp_vendor_codec_t info;
+        uint8_t direction;
+        uint8_t sink_frequency:4;
+        uint8_t source_frequency:4;
+} __attribute__ ((packed)) a2dp_faststream_t;
+
 #elif __BYTE_ORDER == __BIG_ENDIAN
 
 typedef struct {
@@ -287,59 +326,67 @@ typedef struct {
 	uint32_t rfa;
 } __attribute__ ((packed)) a2dp_aptx_hd_t;
 
+typedef struct {
+        a2dp_aptx_t aptx;
+        uint8_t reserved:6;
+        uint8_t has_new_caps:1;
+        uint8_t bidirect_link:1;
+} __attribute__ ((packed)) a2dp_aptx_ll_t;
+
+typedef struct {
+        a2dp_vendor_codec_t info;
+        uint8_t direction;
+        uint8_t source_frequency:4;
+        uint8_t sink_frequency:4;
+} __attribute__ ((packed)) a2dp_faststream_t;
 
 #else
 #error "Unknown byte order"
 #endif
 
-static inline int a2dp_sbc_get_channels(a2dp_sbc_t *config)
-{
-	switch (config->channel_mode) {
-	case SBC_CHANNEL_MODE_MONO:
-                return 1;
-	case SBC_CHANNEL_MODE_DUAL_CHANNEL:
-	case SBC_CHANNEL_MODE_STEREO:
-	case SBC_CHANNEL_MODE_JOINT_STEREO:
-                return 2;
-	default:
-                return -1;
-        }
-}
+typedef struct {
+        a2dp_aptx_ll_t base;
+        uint8_t reserved;
+        uint8_t target_level2;
+        uint8_t target_level1;
+        uint8_t initial_level2;
+        uint8_t initial_level1;
+        uint8_t sra_max_rate;
+        uint8_t sra_avg_time;
+        uint8_t good_working_level2;
+        uint8_t good_working_level1;
+} __attribute__ ((packed)) a2dp_aptx_ll_ext_t;
 
-static inline int a2dp_sbc_get_frequency(a2dp_sbc_t *config)
-{
-	switch (config->frequency) {
-	case SBC_SAMPLING_FREQ_16000:
-                return 16000;
-	case SBC_SAMPLING_FREQ_32000:
-                return 32000;
-	case SBC_SAMPLING_FREQ_44100:
-                return 44100;
-	case SBC_SAMPLING_FREQ_48000:
-                return 48000;
-	default:
-                return -1;
-        }
-}
+
+#define A2DP_CODEC_DEFAULT_RATE		48000
+#define A2DP_CODEC_DEFAULT_CHANNELS	2
+
+struct a2dp_codec_audio_info {
+	uint32_t rate;
+	uint32_t channels;
+};
 
 struct a2dp_codec_handle;
 
 struct a2dp_codec {
+	enum spa_bluetooth_audio_codec id;
 	uint8_t codec_id;
 	a2dp_vendor_codec_t vendor;
 
 	const char *name;
 	const char *description;
+	const char *endpoint_name;	/**< Endpoint name. If NULL, same as name */
 	const struct spa_dict *info;
 
 	const size_t send_buf_size;
 
-	const char *feature_flag;
+	const struct a2dp_codec *duplex_codec;	/**< Codec for non-standard A2DP duplex channel */
 
 	int (*fill_caps) (const struct a2dp_codec *codec, uint32_t flags,
 			uint8_t caps[A2DP_MAX_CAPS_SIZE]);
 	int (*select_config) (const struct a2dp_codec *codec, uint32_t flags,
 			const void *caps, size_t caps_size,
+			const struct a2dp_codec_audio_info *info,
 			const struct spa_dict *settings, uint8_t config[A2DP_MAX_CAPS_SIZE]);
 	int (*enum_config) (const struct a2dp_codec *codec,
 			const void *caps, size_t caps_size, uint32_t id, uint32_t idx,
@@ -354,16 +401,21 @@ struct a2dp_codec {
 	 * otherwise not checked beforehand.
 	 */
 	int (*caps_preference_cmp) (const struct a2dp_codec *codec, const void *caps1, size_t caps1_size,
-			const void *caps2, size_t caps2_size);
+			const void *caps2, size_t caps2_size, const struct a2dp_codec_audio_info *info);
+
+	void *(*init_props) (const struct a2dp_codec *codec, const struct spa_dict *settings);
+	void (*clear_props) (void *);
+	int (*enum_props) (void *props, const struct spa_dict *settings, uint32_t id, uint32_t idx,
+			struct spa_pod_builder *builder, struct spa_pod **param);
+	int (*set_props) (void *props, const struct spa_pod *param);
 
 	void *(*init) (const struct a2dp_codec *codec, uint32_t flags, void *config, size_t config_size,
-			const struct spa_audio_info *info, const struct spa_dict *settings, size_t mtu);
+			const struct spa_audio_info *info, void *props, size_t mtu);
 	void (*deinit) (void *data);
 
-	int (*update_settings) (void *data, const struct spa_dict *settings);
+	int (*update_props) (void *data, void *props);
 
 	int (*get_block_size) (void *data);
-	int (*get_num_blocks) (void *data);
 
 	int (*abr_process) (void *data, size_t unsent);
 
@@ -372,7 +424,7 @@ struct a2dp_codec {
 	int (*encode) (void *data,
 		const void *src, size_t src_size,
 		void *dst, size_t dst_size,
-		size_t *dst_out);
+		size_t *dst_out, int *need_flush);
 
 	int (*start_decode) (void *data,
 		const void *src, size_t src_size, uint16_t *seqnum, uint32_t *timestamp);
@@ -385,8 +437,18 @@ struct a2dp_codec {
 	int (*increase_bitpool) (void *data);
 };
 
-extern const struct a2dp_codec **a2dp_codecs;
+extern const struct a2dp_codec * const * const a2dp_codecs;
 
-bool a2dp_codec_check_caps(const struct a2dp_codec *codec, unsigned int codec_id, const void *caps, size_t caps_size);
+struct a2dp_codec_config {
+	uint32_t config;
+	int value;
+	unsigned int priority;
+};
+
+int a2dp_codec_select_config(const struct a2dp_codec_config configs[], size_t n,
+	uint32_t cap, int preferred_value);
+
+bool a2dp_codec_check_caps(const struct a2dp_codec *codec, unsigned int codec_id,
+	const void *caps, size_t caps_size, const struct a2dp_codec_audio_info *info);
 
 #endif
