@@ -30,10 +30,14 @@
 
 #include "config.h"
 
+#include <spa/utils/string.h>
+
 #include "pipewire/pipewire.h"
 
 #include "media-session.h"
 
+/** \page page_media_session_module_access_flatpak Media Session Module: Access Flatpak
+ */
 #define NAME		"access-flatpak"
 #define SESSION_KEY	"access-flatpak"
 
@@ -71,11 +75,11 @@ static void object_update(void *data)
 
 		if (client->obj->info == NULL || client->obj->info->props == NULL ||
 		    (str = spa_dict_lookup(client->obj->info->props, PW_KEY_ACCESS)) == NULL ||
-		    strcmp(str, "flatpak") != 0)
+		    !spa_streq(str, "flatpak"))
 			return;
 
 		if ((str = spa_dict_lookup(client->obj->info->props, PW_KEY_MEDIA_CATEGORY)) != NULL &&
-		    (strcmp(str, "Manager") == 0)) {
+		    (spa_streq(str, "Manager"))) {
 			/* FIXME, use permission store to check if this app is allowed to
 			 * be a manager app */
 			perms = PW_PERM_ALL;
@@ -131,7 +135,7 @@ static void session_create(void *data, struct sm_object *object)
 
 	pw_log_debug(NAME " %p: create global '%d'", impl, object->id);
 
-	if (strcmp(object->type, PW_TYPE_INTERFACE_Client) == 0)
+	if (spa_streq(object->type, PW_TYPE_INTERFACE_Client))
 		res = handle_client(impl, object);
 	else
 		res = 0;
@@ -145,7 +149,7 @@ static void session_remove(void *data, struct sm_object *object)
 	struct impl *impl = data;
 	pw_log_debug(NAME " %p: remove global '%d'", impl, object->id);
 
-	if (strcmp(object->type, PW_TYPE_INTERFACE_Client) == 0) {
+	if (spa_streq(object->type, PW_TYPE_INTERFACE_Client)) {
 		struct client *client;
 
 		if ((client = sm_object_get_data(object, SESSION_KEY)) != NULL)
