@@ -42,6 +42,7 @@
 #include <spa/utils/names.h>
 #include <spa/utils/keys.h>
 #include <spa/utils/json.h>
+#include <spa/utils/string.h>
 #include <spa/param/props.h>
 #include <spa/pod/builder.h>
 #include <spa/pod/parser.h>
@@ -50,13 +51,15 @@
 #include <spa/support/dbus.h>
 
 #include <pipewire/pipewire.h>
-#include <pipewire/main-loop.h>
-#include <extensions/session-manager.h>
+#include <pipewire/i18n.h>
+#include <pipewire/extensions/session-manager.h>
 
 #include "media-session.h"
 
 #include "reserve.c"
 
+/** \page page_media_session_module_alsa_monitor Media Session Module: ALSA Monitor
+ */
 #define SESSION_CONF	"alsa-monitor.conf"
 
 #define DEFAULT_JACK_SECONDS	1
@@ -117,8 +120,6 @@ struct impl {
 	struct pw_properties *conf;
 	struct pw_properties *props;
 
-	DBusConnection *conn;
-
 	struct spa_handle *handle;
 
 	struct spa_device *monitor;
@@ -128,12 +129,14 @@ struct impl {
 
 	struct spa_source *jack_timeout;
 	struct pw_proxy *jack_device;
+	unsigned int reserve:1;
 };
 
 #undef NAME
 #define NAME "alsa-monitor"
 
 static int probe_device(struct device *device);
+static int do_device_acquire(struct device *device);
 
 static struct node *alsa_find_node(struct device *device, uint32_t id, const char *name)
 {
@@ -145,7 +148,7 @@ static struct node *alsa_find_node(struct device *device, uint32_t id, const cha
 			return node;
 		if (name != NULL &&
 		    (str = pw_properties_get(node->props, PW_KEY_NODE_NAME)) != NULL &&
-		    strcmp(name, str) == 0)
+		    spa_streq(name, str))
 			return node;
 	}
 	return NULL;
@@ -174,8 +177,8 @@ static int node_acquire(void *data)
 
 	node->acquired = true;
 
-	if (device && device->n_acquired++ == 0 && device->reserve)
-		return rd_device_acquire(device->reserve);
+	if (device && device->n_acquired++ == 0)
+		return do_device_acquire(device);
 	else
 		return 0;
 }
@@ -203,6 +206,61 @@ static const struct sm_object_methods node_methods = {
 	.release = node_release,
 };
 
+static void update_icon_name(struct pw_properties *p, bool is_sink)
+{
+	const char *s, *d = NULL, *bus;
+
+	if ((s = pw_properties_get(p, PW_KEY_DEVICE_FORM_FACTOR))) {
+		if (spa_streq(s, "microphone"))
+			d = "audio-input-microphone";
+		else if (spa_streq(s, "webcam"))
+			d = "camera-web";
+		else if (spa_streq(s, "computer"))
+			d = "computer";
+		else if (spa_streq(s, "handset"))
+			d = "phone";
+		else if (spa_streq(s, "portable"))
+			d = "multimedia-player";
+		else if (spa_streq(s, "tv"))
+			d = "video-display";
+		else if (spa_streq(s, "headset"))
+			d = "audio-headset";
+		else if (spa_streq(s, "headphone"))
+			d = "audio-headphones";
+		else if (spa_streq(s, "speaker"))
+			d = "audio-speakers";
+		else if (spa_streq(s, "hands-free"))
+			d = "audio-handsfree";
+	}
+	if (!d)
+		if ((s = pw_properties_get(p, PW_KEY_DEVICE_CLASS)))
+			if (spa_streq(s, "modem"))
+				d = "modem";
+
+	if (!d) {
+		if (is_sink)
+			d = "audio-card";
+		else
+			d = "audio-input-microphone";
+	}
+
+	if ((s = pw_properties_get(p, "device.profile.name")) != NULL) {
+		if (strstr(s, "analog"))
+			s = "-analog";
+		else if (strstr(s, "iec958"))
+			s = "-iec958";
+		else if (strstr(s, "hdmi"))
+			s = "-hdmi";
+		else
+			s = NULL;
+	}
+
+	bus = pw_properties_get(p, PW_KEY_DEVICE_BUS);
+
+	pw_properties_setf(p, PW_KEY_DEVICE_ICON_NAME,
+			"%s%s%s%s", d, s ? s : "", bus ? "-" : "", bus ? bus : "");
+}
+
 static struct node *alsa_create_node(struct device *device, uint32_t id,
 		const struct spa_device_object_info *info)
 {
@@ -217,7 +275,7 @@ static struct node *alsa_create_node(struct device *device, uint32_t id,
 	if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
 		spa_debug_dict(0, info->props);
 
-	if (strcmp(info->type, SPA_TYPE_INTERFACE_Node) != 0) {
+	if (!spa_streq(info->type, SPA_TYPE_INTERFACE_Node)) {
 		errno = EINVAL;
 		return NULL;
 	}
@@ -248,7 +306,7 @@ static struct node *alsa_create_node(struct device *device, uint32_t id,
 		profile = "unknown";
 	profile_desc = pw_properties_get(node->props, "device.profile.description");
 
-	if (!strcmp(stream, "capture"))
+	if (spa_streq(stream, "capture"))
 		node->direction = PW_DIRECTION_OUTPUT;
 	else
 		node->direction = PW_DIRECTION_INPUT;
@@ -265,9 +323,9 @@ static struct node *alsa_create_node(struct device *device, uint32_t id,
 	priority -= atol(dev) * 16;
 	priority -= atol(subdev);
 
-	if (strstr(profile, "analog-") == profile)
+	if (spa_strstartswith(profile, "analog-"))
 		priority += 9;
-	else if (strstr(profile, "iec958-") == profile)
+	else if (spa_strstartswith(profile, "iec958-"))
 		priority += 8;
 
 	if (pw_properties_get(node->props, PW_KEY_PRIORITY_DRIVER) == NULL) {
@@ -300,7 +358,7 @@ static struct node *alsa_create_node(struct device *device, uint32_t id,
 
 		if ((devname = pw_properties_get(device->props, SPA_KEY_DEVICE_NAME)) == NULL)
 			devname = "unnamed-device";
-		if (strstr(devname, "alsa_card.") == devname)
+		if (spa_strstartswith(devname, "alsa_card."))
 			devname += 10;
 
 		pw_properties_set(node->props, SPA_KEY_NODE_NAME,
@@ -339,11 +397,11 @@ static struct node *alsa_create_node(struct device *device, uint32_t id,
 			pw_properties_set(node->props, PW_KEY_NODE_DESCRIPTION,
 				sm_media_session_sanitize_description(tmp, sizeof(tmp),
 					' ', "%s %s", desc, profile_desc));
-		} else if (strcmp(subdev, "0")) {
+		} else if (!spa_streq(subdev, "0")) {
 			pw_properties_set(node->props, PW_KEY_NODE_DESCRIPTION,
 				sm_media_session_sanitize_description(tmp, sizeof(tmp),
 					' ', "%s (%s %s)", desc, name, subdev));
-		} else if (strcmp(dev, "0")) {
+		} else if (!spa_streq(dev, "0")) {
 			pw_properties_set(node->props, PW_KEY_NODE_DESCRIPTION,
 				sm_media_session_sanitize_description(tmp, sizeof(tmp),
 					' ', "%s (%s)", desc, name));
@@ -353,6 +411,8 @@ static struct node *alsa_create_node(struct device *device, uint32_t id,
 					' ', "%s", desc));
 		}
 	}
+	if (pw_properties_get(node->props, PW_KEY_DEVICE_ICON_NAME) == NULL)
+		update_icon_name(node->props, node->direction == PW_DIRECTION_INPUT);
 
 	node->impl = impl;
 	node->device = device;
@@ -467,7 +527,7 @@ static struct device *alsa_find_device(struct impl *impl, uint32_t id, const cha
 			return device;
 		if (name != NULL &&
 		    (str = pw_properties_get(device->props, PW_KEY_DEVICE_NAME)) != NULL &&
-		    strcmp(str, name) == 0)
+		    spa_streq(str, name))
 			return device;
 	}
 	return NULL;
@@ -478,10 +538,12 @@ static void alsa_update_device(struct impl *impl, struct device *device,
 {
 	pw_log_debug("update device %u", device->id);
 
-	if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
-		spa_debug_dict(0, info->props);
+	if (info->change_mask & SPA_DEVICE_OBJECT_CHANGE_MASK_PROPS) {
+		if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
+			spa_debug_dict(0, info->props);
 
-	pw_properties_update(device->props, info->props);
+		pw_properties_update(device->props, info->props);
+	}
 }
 
 static int update_device_props(struct device *device)
@@ -522,12 +584,12 @@ static int update_device_props(struct device *device)
 		d = NULL;
 
 		if ((s = pw_properties_get(p, PW_KEY_DEVICE_FORM_FACTOR)))
-			if (strcmp(s, "internal") == 0)
-				d = "Built-in Audio";
+			if (spa_streq(s, "internal"))
+				d = _("Built-in Audio");
 		if (!d)
 			if ((s = pw_properties_get(p, PW_KEY_DEVICE_CLASS)))
-				if (strcmp(s, "modem") == 0)
-					d = "Modem";
+				if (spa_streq(s, "modem"))
+					d = _("Modem");
 		if (!d)
 			d = pw_properties_get(p, PW_KEY_DEVICE_PRODUCT_NAME);
 
@@ -536,7 +598,7 @@ static int update_device_props(struct device *device)
 		if (!d)
 			d = pw_properties_get(p, "alsa.card_name");
 		if (!d)
-			d = "Unknown device";
+			d = _("Unknown device");
 
 		pw_properties_set(p, PW_KEY_DEVICE_DESCRIPTION, d);
 	}
@@ -547,44 +609,9 @@ static int update_device_props(struct device *device)
 			pw_properties_set(p, PW_KEY_DEVICE_NICK, s);
 	}
 
-	if (pw_properties_get(p, PW_KEY_DEVICE_ICON_NAME) == NULL) {
-		d = NULL;
+	if (pw_properties_get(p, PW_KEY_DEVICE_ICON_NAME) == NULL)
+		update_icon_name(device->props, true);
 
-		if ((s = pw_properties_get(p, PW_KEY_DEVICE_FORM_FACTOR))) {
-			if (strcmp(s, "microphone") == 0)
-				d = "audio-input-microphone";
-			else if (strcmp(s, "webcam") == 0)
-				d = "camera-web";
-			else if (strcmp(s, "computer") == 0)
-				d = "computer";
-			else if (strcmp(s, "handset") == 0)
-				d = "phone";
-			else if (strcmp(s, "portable") == 0)
-				d = "multimedia-player";
-			else if (strcmp(s, "tv") == 0)
-				d = "video-display";
-			else if (strcmp(s, "headset") == 0)
-				d = "audio-headset";
-			else if (strcmp(s, "headphone") == 0)
-				d = "audio-headphones";
-			else if (strcmp(s, "speaker") == 0)
-				d = "audio-speakers";
-			else if (strcmp(s, "hands-free") == 0)
-				d = "audio-handsfree";
-		}
-		if (!d)
-			if ((s = pw_properties_get(p, PW_KEY_DEVICE_CLASS)))
-				if (strcmp(s, "modem") == 0)
-					d = "modem";
-
-		if (!d)
-			d = "audio-card";
-
-		s = pw_properties_get(p, PW_KEY_DEVICE_BUS);
-
-		pw_properties_setf(p, PW_KEY_DEVICE_ICON_NAME,
-				"%s-analog%s%s", d, s ? "-" : "", s);
-	}
 	return 1;
 }
 
@@ -653,19 +680,22 @@ static void add_jack_timeout(struct impl *impl)
 	pw_loop_update_timer(main_loop, impl->jack_timeout, &value, NULL, false);
 }
 
-static void reserve_acquired(void *data, struct rd_device *d)
+static void do_reserve_acquired(struct device *device)
 {
-	struct device *device = data;
-
-	pw_log_info("%p: reserve acquired %d", device, device->n_acquired);
-
 	if (!device->probed)
 		probe_device(device);
 
-	if (device->n_acquired == 0)
+	if (device->n_acquired == 0 && device->reserve)
 		rd_device_release(device->reserve);
 }
 
+static void reserve_acquired(void *data, struct rd_device *d)
+{
+	struct device *device = data;
+	pw_log_info("%p: reserve acquired %d", device, device->n_acquired);
+	do_reserve_acquired(device);
+}
+
 static void complete_release(struct device *device)
 {
 	if (device->reserve)
@@ -729,7 +759,7 @@ static void reserve_busy(void *data, struct rd_device *d, const char *name, int3
 
 	device->sdevice->locked = true;
 
-	if (strcmp(name, "jack") == 0) {
+	if (spa_streq(name, "jack")) {
 		add_jack_timeout(impl);
 	} else {
 		remove_jack_timeout(impl);
@@ -748,10 +778,9 @@ static void reserve_available(void *data, struct rd_device *d, const char *name)
 	device->sdevice->locked = false;
 
 	remove_jack_timeout(impl);
-	if (strcmp(name, "jack") == 0) {
+	if (spa_streq(name, "jack")) {
 		set_jack_profile(impl, 0);
 	}
-
 }
 
 static const struct rd_device_callbacks reserve_callbacks = {
@@ -761,6 +790,61 @@ static const struct rd_device_callbacks reserve_callbacks = {
 	.available = reserve_available,
 };
 
+static int do_device_acquire(struct device *device)
+{
+	struct impl *impl = device->impl;
+	struct sm_media_session *session = impl->session;
+	int res;
+
+	if (!impl->reserve)
+		goto done;
+
+	if (device->reserve == NULL) {
+		const char *reserve;
+		const char *path = pw_properties_get(device->props, SPA_KEY_API_ALSA_PATH);
+		DBusConnection *conn = NULL;
+
+		if (session->dbus_connection)
+			conn = spa_dbus_connection_get(session->dbus_connection);
+		if (conn == NULL) {
+			pw_log_warn("no dbus connection, device reservation disabled");
+			goto done;
+		}
+		pw_log_debug("got dbus connection %p", conn);
+
+		reserve = pw_properties_get(device->props, "api.dbus.ReserveDevice1");
+		if (reserve == NULL)
+			goto done;
+
+		device->reserve = rd_device_new(conn, reserve,
+				"PipeWire", -10,
+				&reserve_callbacks, device);
+
+		if (device->reserve == NULL) {
+			pw_log_warn("can't create device reserve for %s: %m", reserve);
+			goto done;
+		} else if (path) {
+			rd_device_set_application_device_name(device->reserve, path);
+		} else {
+			pw_log_warn("empty reserve device path for %s", reserve);
+		}
+	}
+	res = rd_device_acquire(device->reserve);
+	if (res < 0 && res != -EBUSY) {
+		pw_log_warn("device reserve failed, disabling: %s", spa_strerror(res));
+		goto done;
+	}
+	return res;
+
+done:
+	do_reserve_acquired(device);
+	if (device->reserve != NULL) {
+		rd_device_destroy(device->reserve);
+		device->reserve = NULL;
+	}
+	return 0;
+}
+
 static void device_destroy(void *data)
 {
 	struct device *device = data;
@@ -826,6 +910,9 @@ static int probe_device(struct device *device)
 	void *iface;
 	int res;
 
+	if (device->probed)
+		return 0;
+
 	handle = pw_context_load_spa_handle(context,
 			device->factory_name, &device->props->dict);
 	if (handle == NULL) {
@@ -874,7 +961,7 @@ static struct device *alsa_create_device(struct impl *impl, uint32_t id,
 	if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
 		spa_debug_dict(0, info->props);
 
-	if (strcmp(info->type, SPA_TYPE_INTERFACE_Device) != 0) {
+	if (!spa_streq(info->type, SPA_TYPE_INTERFACE_Device)) {
 		errno = EINVAL;
 		return NULL;
 	}
@@ -905,30 +992,11 @@ static struct device *alsa_create_device(struct impl *impl, uint32_t id,
 	else
 		device->factory_name = strdup(info->factory_name);
 
-	if (impl->conn &&
-	    (card = spa_dict_lookup(info->props, SPA_KEY_API_ALSA_CARD)) != NULL) {
-		const char *reserve;
-
+	if ((card = spa_dict_lookup(info->props, SPA_KEY_API_ALSA_CARD)) != NULL) {
 		device->priority -= atol(card) * 64;
-
 		pw_properties_setf(device->props, "api.dbus.ReserveDevice1", "Audio%s", card);
-		reserve = pw_properties_get(device->props, "api.dbus.ReserveDevice1");
-
-		device->reserve = rd_device_new(impl->conn, reserve,
-				"PipeWire", -10,
-				&reserve_callbacks, device);
-
-		if (device->reserve == NULL) {
-			pw_log_warn("can't create device reserve for %s: %m", reserve);
-		} else {
-			rd_device_set_application_device_name(device->reserve,
-				spa_dict_lookup(info->props, SPA_KEY_API_ALSA_PATH));
-		}
 	}
-	if (device->reserve != NULL)
-		rd_device_acquire(device->reserve);
-	else
-		probe_device(device);
+	do_device_acquire(device);
 
 	return device;
 exit:
@@ -1011,9 +1079,22 @@ static void session_destroy(void *data)
 	free(impl);
 }
 
+static void session_dbus_disconnected(void *data)
+{
+	struct device *d;
+	struct impl *impl = data;
+
+	spa_list_for_each(d, &impl->device_list, link) {
+		if (d->reserve != NULL)
+			rd_device_destroy(d->reserve);
+		d->reserve = NULL;
+	}
+}
+
 static const struct sm_media_session_events session_events = {
 	SM_VERSION_MEDIA_SESSION_EVENTS,
 	.destroy = session_destroy,
+	.dbus_disconnected = session_dbus_disconnected,
 };
 
 int sm_alsa_monitor_start(struct sm_media_session *session)
@@ -1047,18 +1128,13 @@ int sm_alsa_monitor_start(struct sm_media_session *session)
 		pw_properties_update_string(impl->props, str, strlen(str));
 
 	if ((str = pw_properties_get(impl->props, "alsa.reserve")) == NULL ||
-	    pw_properties_parse_bool(str)) {
-		if (session->dbus_connection)
-			impl->conn = spa_dbus_connection_get(session->dbus_connection);
-		if (impl->conn == NULL)
-			pw_log_warn("no dbus connection, device reservation disabled");
-		else
-			pw_log_debug("got dbus connection %p", impl->conn);
-	}
+	    pw_properties_parse_bool(str))
+		impl->reserve = true;
 
 	impl->handle = pw_context_load_spa_handle(context, SPA_NAME_API_ALSA_ENUM_UDEV, NULL);
 	if (impl->handle == NULL) {
 		res = -errno;
+		pw_log_info("can't load %s: %m", SPA_NAME_API_ALSA_ENUM_UDEV);
 		goto out_free;
 	}
 
@@ -1083,10 +1159,8 @@ int sm_alsa_monitor_start(struct sm_media_session *session)
 out_free:
 	if (impl->handle)
 		pw_unload_spa_handle(impl->handle);
-	if (impl->conf)
-		pw_properties_free(impl->conf);
-	if (impl->props)
-		pw_properties_free(impl->props);
+	pw_properties_free(impl->conf);
+	pw_properties_free(impl->props);
 	free(impl);
 	return res;
 }
