@@ -35,15 +35,18 @@
 #include <spa/utils/hook.h>
 #include <spa/utils/result.h>
 #include <spa/utils/json.h>
+#include <spa/utils/string.h>
 #include <spa/pod/parser.h>
 #include <spa/pod/builder.h>
 #include <spa/debug/pod.h>
 
 #include "pipewire/pipewire.h"
-#include "extensions/metadata.h"
+#include "pipewire/extensions/metadata.h"
 
 #include "media-session.h"
 
+/** \page page_media_session_module_default_routes Media Session Module: Default Routes
+ */
 #define NAME		"default-routes"
 #define SESSION_KEY	"default-routes"
 #define PREFIX		"default.route."
@@ -117,7 +120,7 @@ static uint32_t channel_from_name(const char *name)
 {
 	int i;
 	for (i = 0; spa_type_audio_channel[i].name; i++) {
-		if (strcmp(name, spa_debug_type_short_name(spa_type_audio_channel[i].name)) == 0)
+		if (spa_streq(name, spa_debug_type_short_name(spa_type_audio_channel[i].name)))
 			return spa_type_audio_channel[i].type;
 	}
 	return SPA_AUDIO_CHANNEL_UNKNOWN;
@@ -136,8 +139,8 @@ static const char *channel_to_name(uint32_t channel)
 struct route_info {
 	uint32_t index;
 	uint32_t generation;
-	uint32_t available;
-	uint32_t prev_available;
+	enum spa_param_availability available;
+	enum spa_param_availability prev_available;
 	enum spa_direction direction;
 	char name[64];
 	unsigned int save:1;
@@ -152,8 +155,9 @@ struct route {
 	enum spa_direction direction;
 	const char *name;
 	uint32_t priority;
-	uint32_t available;
+	enum spa_param_availability available;
 	struct spa_pod *props;
+	struct spa_pod *profiles;
 	bool save;
 };
 
@@ -230,7 +234,8 @@ static int parse_enum_route(struct sm_param *p, uint32_t device_id, struct route
 			SPA_PARAM_ROUTE_name, SPA_POD_String(&r->name),
 			SPA_PARAM_ROUTE_priority,  SPA_POD_OPT_Int(&r->priority),
 			SPA_PARAM_ROUTE_available,  SPA_POD_OPT_Id(&r->available),
-			SPA_PARAM_ROUTE_devices, SPA_POD_OPT_Pod(&devices))) < 0)
+			SPA_PARAM_ROUTE_devices, SPA_POD_OPT_Pod(&devices),
+			SPA_PARAM_ROUTE_profiles, SPA_POD_OPT_Pod(&r->profiles))) < 0)
 		return res;
 
 	if (device_id != SPA_ID_INVALID && !array_contains(devices, device_id))
@@ -345,21 +350,21 @@ static int restore_route_params(struct device *dev, const char *val, struct rout
 			SPA_TYPE_OBJECT_Props, SPA_PARAM_Route);
 
 	while (spa_json_get_string(&it[1], key, sizeof(key)-1) > 0) {
-		if (strcmp(key, "volume") == 0) {
+		if (spa_streq(key, "volume")) {
 			float vol;
 			if (spa_json_get_float(&it[1], &vol) <= 0)
                                 continue;
 			spa_pod_builder_prop(&b, SPA_PROP_volume, 0);
 			spa_pod_builder_float(&b, vol);
 		}
-		else if (strcmp(key, "mute") == 0) {
+		else if (spa_streq(key, "mute")) {
 			bool mute;
 			if (spa_json_get_bool(&it[1], &mute) <= 0)
                                 continue;
 			spa_pod_builder_prop(&b, SPA_PROP_mute, 0);
 			spa_pod_builder_bool(&b, mute);
 		}
-		else if (strcmp(key, "volumes") == 0) {
+		else if (spa_streq(key, "volumes")) {
 			uint32_t n_vols;
 			float vols[SPA_AUDIO_MAX_CHANNELS];
 
@@ -377,7 +382,7 @@ static int restore_route_params(struct device *dev, const char *val, struct rout
 			spa_pod_builder_array(&b, sizeof(float), SPA_TYPE_Float,
 					n_vols, vols);
 		}
-		else if (strcmp(key, "channels") == 0) {
+		else if (spa_streq(key, "channels")) {
 			uint32_t n_ch;
 			uint32_t map[SPA_AUDIO_MAX_CHANNELS];
 
@@ -397,7 +402,7 @@ static int restore_route_params(struct device *dev, const char *val, struct rout
 			spa_pod_builder_array(&b, sizeof(uint32_t), SPA_TYPE_Id,
 					n_ch, map);
 		}
-		else if (strcmp(key, "latencyOffsetNsec") == 0) {
+		else if (spa_streq(key, "latencyOffsetNsec")) {
 			float delay;
 			if (spa_json_get_float(&it[1], &delay) <= 0)
                                 continue;
@@ -565,13 +570,29 @@ static int find_best_route(struct device *dev, uint32_t device_id, struct route
 		    parse_enum_route(p, device_id, &t) < 0)
 			continue;
 
-		if (t.available == SPA_PARAM_AVAILABILITY_yes) {
-			if (best_avail.name == NULL || t.priority > best_avail.priority)
+		if (t.available == SPA_PARAM_AVAILABILITY_yes || t.available == SPA_PARAM_AVAILABILITY_unknown) {
+			struct route_info *ri;
+			if ((ri = find_route_info(dev, &t)) && ri->direction == SPA_DIRECTION_OUTPUT &&
+			    ri->available != ri->prev_available) {
+				/* If route availability changed, that means a user just
+				 * plugged in something like headphones, and they probably
+				 * expect to hear sound from it. Switch to it immediately.
+				 *
+				 * TODO: switch INPUT ports without source and the input
+				 * ports their source->active_port is part of a group of
+				 * ports (see module-switch-on-port-available.c in PulseAudio).
+				 */
 				best_avail = t;
-		}
-		else if (t.available != SPA_PARAM_AVAILABILITY_no) {
-			if (best_unk.name == NULL || t.priority > best_unk.priority)
-				best_unk = t;
+				ri->save = true;
+				break;
+			}
+			else if (t.available == SPA_PARAM_AVAILABILITY_yes) {
+				if (best_avail.name == NULL || t.priority > best_avail.priority)
+					best_avail = t;
+			} else { // SPA_PARAM_AVAILABILITY_unknown
+				if (best_unk.name == NULL || t.priority > best_unk.priority)
+					best_unk = t;
+			}
 		}
 	}
 	best = best_avail;
@@ -591,7 +612,7 @@ static int find_route(struct device *dev, uint32_t device_id, const char *name,
 		if (p->id != SPA_PARAM_EnumRoute ||
 		    parse_enum_route(p, device_id, r) < 0)
 			continue;
-		if (strcmp(r->name, name) != 0)
+		if (!spa_streq(r->name, name))
 			continue;
 		return 0;
 	}
@@ -689,7 +710,7 @@ static int reconfigure_profile(struct device *dev, struct profile *pr, bool rest
 						NULL) >= 0) {
 				if (key == NULL || val == NULL)
 					break;
-				if (strcmp(key, "card.profile.devices") == 0) {
+				if (spa_streq(key, "card.profile.devices")) {
 					uint32_t *devices, n_devices, i;
 
 					devices = spa_pod_get_array(val, &n_devices);
@@ -733,10 +754,10 @@ static int handle_route(struct device *dev, struct route *r)
 
 	if (!ri->prev_active) {
 		/* a new port has been found, restore the volume and make sure we
-		 * save this as a prefered port */
+		 * save this as a preferred port */
 		pw_log_info("device %d: new active port found '%s'", dev->id, r->name);
 		restore_route(dev, r);
-	} else if (r->props) {
+	} else if (r->props && r->save) {
 		/* just save port properties */
 		save_route(dev, r);
 	}
@@ -747,11 +768,13 @@ static int handle_device(struct device *dev)
 {
 	struct profile pr;
 	struct sm_param *p;
-	int res;
 	bool route_changed = false;
 
 	dev->generation++;
 
+	if (find_current_profile(dev, &pr) < 0)
+		pr.index = SPA_ID_INVALID;
+
 	/* first look at all routes and update */
 	spa_list_for_each(p, &dev->obj->param_list, link) {
 		struct route r;
@@ -769,7 +792,8 @@ static int handle_device(struct device *dev)
 			pw_log_info("device %d: route %s available changed %d -> %d",
 					dev->id, r.name, ri->available, r.available);
 			ri->available = r.available;
-			route_changed = true;
+			if (array_contains(r.profiles, pr.index))
+				route_changed = true;
 		}
 		ri->generation = dev->generation;
 		ri->prev_active = ri->active;
@@ -787,7 +811,7 @@ static int handle_device(struct device *dev)
 
 	prune_route_info(dev);
 
-	if ((res = find_current_profile(dev, &pr)) >= 0) {
+	if (pr.index != SPA_ID_INVALID) {
 		bool restore = dev->active_profile != pr.index;
 		if (restore || route_changed)
 			reconfigure_profile(dev, &pr, restore);
@@ -820,7 +844,7 @@ static void session_create(void *data, struct sm_object *object)
 	struct device *dev;
 	const char *name;
 
-	if (strcmp(object->type, PW_TYPE_INTERFACE_Device) != 0 ||
+	if (!spa_streq(object->type, PW_TYPE_INTERFACE_Device) ||
 	    object->props == NULL ||
 	    (name = pw_properties_get(object->props, PW_KEY_DEVICE_NAME)) == NULL)
 		return;
@@ -853,7 +877,7 @@ static void session_remove(void *data, struct sm_object *object)
 	struct impl *impl = data;
 	struct device *dev;
 
-	if (strcmp(object->type, PW_TYPE_INTERFACE_Device) != 0)
+	if (!spa_streq(object->type, PW_TYPE_INTERFACE_Device))
 		return;
 
 	pw_log_debug(NAME " %p: remove device '%d'", impl, object->id);
