@@ -45,6 +45,7 @@
 #include <spa/utils/hook.h>
 #include <spa/utils/result.h>
 #include <spa/utils/json.h>
+#include <spa/utils/string.h>
 #include <spa/param/audio/format-utils.h>
 #include <spa/param/props.h>
 #include <spa/debug/pod.h>
@@ -54,8 +55,8 @@
 #include "pipewire/pipewire.h"
 #include "pipewire/private.h"
 #include "pipewire/conf.h"
-#include "extensions/session-manager.h"
-#include "extensions/client-node.h"
+#include "pipewire/extensions/session-manager.h"
+#include "pipewire/extensions/client-node.h"
 
 #include <dbus/dbus.h>
 
@@ -80,6 +81,7 @@
 #define sm_media_session_emit_shutdown(s)		sm_media_session_emit(s, shutdown, 0)
 #define sm_media_session_emit_destroy(s)		sm_media_session_emit(s, destroy, 0)
 #define sm_media_session_emit_seat_active(s,...)	sm_media_session_emit(s, seat_active, 0, __VA_ARGS__)
+#define sm_media_session_emit_dbus_disconnected(s)	sm_media_session_emit(s, dbus_disconnected, 0)
 
 int sm_access_flatpak_start(struct sm_media_session *sess);
 int sm_access_portal_start(struct sm_media_session *sess);
@@ -92,6 +94,7 @@ int sm_alsa_midi_start(struct sm_media_session *sess);
 int sm_v4l2_monitor_start(struct sm_media_session *sess);
 int sm_libcamera_monitor_start(struct sm_media_session *sess);
 int sm_bluez5_monitor_start(struct sm_media_session *sess);
+int sm_bluez5_autoswitch_start(struct sm_media_session *sess);
 int sm_alsa_monitor_start(struct sm_media_session *sess);
 int sm_suspend_node_start(struct sm_media_session *sess);
 #ifdef HAVE_SYSTEMD
@@ -128,6 +131,7 @@ struct impl {
 
 	struct pw_main_loop *loop;
 	struct spa_dbus *dbus;
+	struct spa_hook dbus_connection_listener;
 
 	struct pw_core *monitor_core;
 	struct spa_hook monitor_listener;
@@ -239,7 +243,7 @@ static void *find_object(struct impl *impl, uint32_t id, const char *type)
 	struct sm_object *obj;
 	if ((obj = pw_map_lookup(&impl->globals, id)) == NULL)
 		return NULL;
-	if (type != NULL && strcmp(obj->type, type) != 0)
+	if (type != NULL && !spa_streq(obj->type, type))
 		return NULL;
 	return obj;
 }
@@ -248,7 +252,7 @@ static struct data *object_find_data(struct sm_object *obj, const char *id)
 {
 	struct data *d;
 	spa_list_for_each(d, &obj->data, link) {
-		if (strcmp(d->id, id) == 0)
+		if (spa_streq(d->id, id))
 			return d;
 	}
 	return NULL;
@@ -271,7 +275,7 @@ void *sm_object_add_data(struct sm_object *obj, const char *id, size_t size)
 
 	spa_list_append(&obj->data, &d->link);
 done:
-	return SPA_MEMBER(d, sizeof(struct data), void);
+	return SPA_PTROFF(d, sizeof(struct data), void);
 }
 
 void *sm_object_get_data(struct sm_object *obj, const char *id)
@@ -280,7 +284,7 @@ void *sm_object_get_data(struct sm_object *obj, const char *id)
 	d = object_find_data(obj, id);
 	if (d == NULL)
 		return NULL;
-	return SPA_MEMBER(d, sizeof(struct data), void);
+	return SPA_PTROFF(d, sizeof(struct data), void);
 }
 
 int sm_object_remove_data(struct sm_object *obj, const char *id)
@@ -338,10 +342,8 @@ unref:
 	if (!obj->discarded)
 		return 0;
 
-	if (obj->props) {
-		pw_properties_free(obj->props);
-		obj->props = NULL;
-	}
+	pw_properties_free(obj->props);
+	obj->props = NULL;
 
 	spa_list_consume(d, &obj->data, link) {
 		spa_list_remove(&d->link);
@@ -361,7 +363,7 @@ int sm_object_destroy(struct sm_object *obj)
 }
 
 static struct param *add_param(struct spa_list *param_list,
-		uint32_t id, const struct spa_pod *param)
+		int seq, int *param_seq, uint32_t id, const struct spa_pod *param)
 {
 	struct param *p;
 
@@ -372,12 +374,25 @@ static struct param *add_param(struct spa_list *param_list,
 	if (id == SPA_ID_INVALID)
 		id = SPA_POD_OBJECT_ID(param);
 
+	if (id >= SM_MAX_PARAMS) {
+		pw_log_error(NAME": too big param id %d", id);
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (seq != param_seq[id]) {
+		pw_log_debug(NAME": ignoring param %d, seq:%d != current_seq:%d",
+				id, seq, param_seq[id]);
+		errno = EBUSY;
+		return NULL;
+	}
+
 	p = malloc(sizeof(struct param) + SPA_POD_SIZE(param));
 	if (p == NULL)
 		return NULL;
 
 	p->this.id = id;
-	p->this.param = SPA_MEMBER(p, sizeof(struct param), struct spa_pod);
+	p->this.param = SPA_PTROFF(p, sizeof(struct param), struct spa_pod);
 	memcpy(p->this.param, param, SPA_POD_SIZE(param));
 
 	spa_list_append(param_list, &p->this.link);
@@ -490,13 +505,21 @@ static void device_event_info(void *object, const struct pw_device_info *info)
 			if (info->params[i].user == 0)
 				continue;
 
+			if (id >= SM_MAX_PARAMS) {
+				pw_log_error(NAME" %p: too big param id %d", impl, id);
+				continue;
+			}
+
 			device->n_params -= clear_params(&device->param_list, id);
 
 			if (info->params[i].flags & SPA_PARAM_INFO_READ) {
-				pw_log_debug(NAME" %p: device %d enum params %d", impl,
-						device->obj.id, id);
-				pw_device_enum_params((struct pw_device*)device->obj.proxy,
-						1, id, 0, UINT32_MAX, NULL);
+				int res;
+				res = pw_device_enum_params((struct pw_device*)device->obj.proxy,
+						++device->param_seq[id], id, 0, UINT32_MAX, NULL);
+				if (SPA_RESULT_IS_ASYNC(res))
+					device->param_seq[id] = res;
+				pw_log_debug(NAME" %p: device %d enum params %d seq:%d", impl,
+						device->obj.id, id, device->param_seq[id]);
 			}
 			info->params[i].user = 0;
 		}
@@ -512,8 +535,8 @@ static void device_event_param(void *object, int seq,
 	struct sm_device *device = object;
 	struct impl *impl = SPA_CONTAINER_OF(device->obj.session, struct impl, this);
 
-	pw_log_debug(NAME" %p: device %p param %d index:%d", impl, device, id, index);
-	if (add_param(&device->param_list, id, param) != NULL)
+	pw_log_debug(NAME" %p: device %p param %d index:%d seq:%d", impl, device, id, index, seq);
+	if (add_param(&device->param_list, seq, device->param_seq, id, param) != NULL)
 		device->n_params++;
 
 	device->obj.avail |= SM_DEVICE_CHANGE_MASK_PARAMS;
@@ -591,13 +614,21 @@ static void node_event_info(void *object, const struct pw_node_info *info)
 			if (info->params[i].user == 0)
 				continue;
 
+			if (id >= SM_MAX_PARAMS) {
+				pw_log_error(NAME" %p: too big param id %d", impl, id);
+				continue;
+			}
+
 			node->n_params -= clear_params(&node->param_list, id);
 
 			if (info->params[i].flags & SPA_PARAM_INFO_READ) {
-				pw_log_debug(NAME" %p: node %d enum params %d", impl,
-						node->obj.id, id);
-				pw_node_enum_params((struct pw_node*)node->obj.proxy,
-						1, id, 0, UINT32_MAX, NULL);
+				int res;
+				res = pw_node_enum_params((struct pw_node*)node->obj.proxy,
+						++node->param_seq[id], id, 0, UINT32_MAX, NULL);
+				if (SPA_RESULT_IS_ASYNC(res))
+					node->param_seq[id] = res;
+				pw_log_debug(NAME" %p: node %d enum params %d seq:%d", impl,
+						node->obj.id, id, node->param_seq[id]);
 			}
 			info->params[i].user = 0;
 		}
@@ -613,8 +644,8 @@ static void node_event_param(void *object, int seq,
 	struct sm_node *node = object;
 	struct impl *impl = SPA_CONTAINER_OF(node->obj.session, struct impl, this);
 
-	pw_log_debug(NAME" %p: node %p param %d index:%d", impl, node, id, index);
-	if (add_param(&node->param_list, id, param) != NULL)
+	pw_log_debug(NAME" %p: node %p param %d index:%d seq:%d", impl, node, id, index, seq);
+	if (add_param(&node->param_list, seq, node->param_seq, id, param) != NULL)
 		node->n_params++;
 
 	node->obj.avail |= SM_NODE_CHANGE_MASK_PARAMS;
@@ -709,7 +740,7 @@ static enum spa_audio_channel find_channel(const char *name)
 {
         int i;
         for (i = 0; spa_type_audio_channel[i].name; i++) {
-                if (strcmp(name, spa_debug_type_short_name(spa_type_audio_channel[i].name)) == 0)
+                if (spa_streq(name, spa_debug_type_short_name(spa_type_audio_channel[i].name)))
                         return spa_type_audio_channel[i].type;
         }
         return SPA_AUDIO_CHANNEL_UNKNOWN;
@@ -724,12 +755,12 @@ static int port_init(void *object)
 
 	if (props) {
 		if ((str = pw_properties_get(props, PW_KEY_PORT_DIRECTION)) != NULL)
-			port->direction = strcmp(str, "out") == 0 ?
+			port->direction = spa_streq(str, "out") ?
 				PW_DIRECTION_OUTPUT : PW_DIRECTION_INPUT;
 		if ((str = pw_properties_get(props, PW_KEY_FORMAT_DSP)) != NULL) {
-			if (strcmp(str, "32 bit float mono audio") == 0)
+			if (spa_streq(str, "32 bit float mono audio"))
 				port->type = SM_PORT_TYPE_DSP_AUDIO;
-			else if (strcmp(str, "8 bit raw midi") == 0)
+			else if (spa_streq(str, "8 bit raw midi"))
 				port->type = SM_PORT_TYPE_DSP_MIDI;
 		}
 		if ((str = pw_properties_get(props, PW_KEY_AUDIO_CHANNEL)) != NULL)
@@ -786,8 +817,7 @@ static void session_event_info(void *object, const struct pw_session_info *info)
 	if (info) {
 		i->change_mask = info->change_mask;
 		if (info->change_mask & PW_SESSION_CHANGE_MASK_PROPS) {
-			if (i->props)
-				pw_properties_free ((struct pw_properties *)i->props);
+			pw_properties_free ((struct pw_properties *)i->props);
 			i->props = (struct spa_dict *) pw_properties_new_dict (info->props);
 		}
 	}
@@ -825,8 +855,7 @@ static void session_destroy(void *object)
 		spa_list_remove(&endpoint->link);
 	}
 	if (i) {
-		if (i->props)
-			pw_properties_free ((struct pw_properties *)i->props);
+		pw_properties_free ((struct pw_properties *)i->props);
 		free(i);
 	}
 
@@ -866,8 +895,7 @@ static void endpoint_event_info(void *object, const struct pw_endpoint_info *inf
 			i->session_id = info->session_id;
 		}
 		if (info->change_mask & PW_ENDPOINT_CHANGE_MASK_PROPS) {
-			if (i->props)
-				pw_properties_free ((struct pw_properties *)i->props);
+			pw_properties_free ((struct pw_properties *)i->props);
 			i->props = (struct spa_dict *) pw_properties_new_dict (info->props);
 			if ((str = spa_dict_lookup(i->props, PW_KEY_PRIORITY_SESSION)) != NULL)
 				endpoint->priority = pw_properties_parse_int(str);
@@ -922,8 +950,7 @@ static void endpoint_destroy(void *object)
 		spa_list_remove(&endpoint->link);
 	}
 	if (i) {
-		if (i->props)
-			pw_properties_free ((struct pw_properties *)i->props);
+		pw_properties_free ((struct pw_properties *)i->props);
 		free(i->name);
 		free(i->media_class);
 		free(i);
@@ -1144,29 +1171,29 @@ static const struct object_info *get_object_info(struct impl *impl, const char *
 {
 	const struct object_info *info;
 
-	if (strcmp(type, PW_TYPE_INTERFACE_Core) == 0)
+	if (spa_streq(type, PW_TYPE_INTERFACE_Core))
 		info = &core_object_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_Module) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Module))
 		info = &module_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_Factory) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Factory))
 		info = &factory_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_Client) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Client))
 		info = &client_info;
-	else if (strcmp(type, SPA_TYPE_INTERFACE_Device) == 0)
+	else if (spa_streq(type, SPA_TYPE_INTERFACE_Device))
 		info = &spa_device_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_Device) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Device))
 		info = &device_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_Node) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Node))
 		info = &node_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_Port) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Port))
 		info = &port_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_Session) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Session))
 		info = &session_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_Endpoint) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_Endpoint))
 		info = &endpoint_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_EndpointStream) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_EndpointStream))
 		info = &endpoint_stream_info;
-	else if (strcmp(type, PW_TYPE_INTERFACE_EndpointLink) == 0)
+	else if (spa_streq(type, PW_TYPE_INTERFACE_EndpointLink))
 		info = &endpoint_link_info;
 	else
 		info = NULL;
@@ -1220,7 +1247,7 @@ create_object(struct impl *impl, struct pw_proxy *proxy, struct pw_proxy *handle
 
 	type = pw_proxy_get_type(handle, NULL);
 
-	if (strcmp(type, PW_TYPE_INTERFACE_ClientNode) == 0)
+	if (spa_streq(type, PW_TYPE_INTERFACE_ClientNode))
 		type = PW_TYPE_INTERFACE_Node;
 
 	info = get_object_info(impl, type);
@@ -1289,8 +1316,7 @@ static void registry_event_free(struct registry_event *re)
 {
 	if (re->proxy)
 		pw_proxy_destroy(re->proxy);
-	if (re->props_store)
-		pw_properties_free(re->props_store);
+	pw_properties_free(re->props_store);
 	if (re->allocated) {
 		spa_list_remove(&re->link);
 		free(re);
@@ -1323,7 +1349,7 @@ static int handle_registry_event(struct impl *impl, struct registry_event *re)
 		 * removed. In that case, we create a zombie object here, but its remove
 		 * event is already queued and arrives soon.
 		 */
-		obj = bind_object(impl, info, re);
+		bind_object(impl, info, re);
 	} else if (obj != NULL && obj->monitor_global == re->monitor) {
 		/* Each core handles their own object updates */
 		update_object(impl, info, obj, re);
@@ -1455,7 +1481,7 @@ monitor_registry_global(void *data, uint32_t id,
 	re.proxy = pw_registry_bind(impl->registry, id, type, info->version, 0);
 	if (re.proxy)
 		handle_registry_event(impl, &re);
-	else 
+	else
 		pw_log_warn(NAME" %p: can't handle global %d: %s", impl, id, spa_strerror(-errno));
 
 	registry_event_free(&re);
@@ -1802,12 +1828,20 @@ static int link_nodes(struct impl *impl, struct endpoint_link *link,
 	struct pw_properties *props;
 	struct sm_port *outport, *inport;
 	int count = 0;
+	bool passive = false;
+	const char *str;
 
 	pw_log_debug(NAME" %p: linking %d -> %d", impl, outnode->obj.id, innode->obj.id);
 
+	if ((str = spa_dict_lookup(outnode->info->props, PW_KEY_NODE_PASSIVE)) != NULL)
+		passive |= (pw_properties_parse_bool(str) || spa_streq(str, "out"));
+	if ((str = spa_dict_lookup(innode->info->props, PW_KEY_NODE_PASSIVE)) != NULL)
+		passive |= (pw_properties_parse_bool(str) || spa_streq(str, "in"));
+
 	props = pw_properties_new(NULL, NULL);
 	pw_properties_setf(props, PW_KEY_LINK_OUTPUT_NODE, "%d", outnode->obj.id);
 	pw_properties_setf(props, PW_KEY_LINK_INPUT_NODE, "%d", innode->obj.id);
+	pw_properties_setf(props, PW_KEY_LINK_PASSIVE, "%s", passive ? "true" : "false");
 
 	spa_list_for_each(inport, &innode->port_list, link)
 		inport->visited = false;
@@ -1998,12 +2032,15 @@ char *sm_media_session_sanitize_name(char *name, int size, char sub, const char
 {
 	char *p;
 	va_list varargs;
+	int res;
 
 	va_start(varargs, fmt);
-	if (vsnprintf(name, size, fmt, varargs) < 0)
-		return NULL;
+	res = vsnprintf(name, size, fmt, varargs);
 	va_end(varargs);
 
+	if (res < 0)
+		return NULL;
+
 	for (p = name; *p; p++) {
 		switch(*p) {
 		case '0' ... '9':
@@ -2023,12 +2060,15 @@ char *sm_media_session_sanitize_description(char *name, int size, char sub, cons
 {
 	char *p;
 	va_list varargs;
+	int res;
 
 	va_start(varargs, fmt);
-	if (vsnprintf(name, size, fmt, varargs) < 0)
-		return NULL;
+	res = vsnprintf(name, size, fmt, varargs);
 	va_end(varargs);
 
+	if (res < 0)
+		return NULL;
+
 	for (p = name; *p; p++) {
 		switch(*p) {
 		case ':':
@@ -2270,6 +2310,18 @@ static int sm_pulse_bridge_start(struct sm_media_session *sess)
 	return 0;
 }
 
+static void dbus_connection_disconnected(void *data)
+{
+	struct impl *impl = data;
+	pw_log_info("DBus disconnected");
+	sm_media_session_emit_dbus_disconnected(impl);
+}
+
+static const struct spa_dbus_connection_events dbus_connection_events = {
+	SPA_VERSION_DBUS_CONNECTION_EVENTS,
+	.disconnected = dbus_connection_disconnected
+};
+
 static void do_quit(void *data, int signal_number)
 {
 	struct impl *impl = data;
@@ -2286,7 +2338,7 @@ static int collect_modules(struct impl *impl, const char *str)
 	int count = 0;
 
 	if ((dir = getenv("PIPEWIRE_CONFIG_DIR")) == NULL)
-		dir = PIPEWIRE_CONFIG_DIR;
+		dir = PIPEWIRE_CONFDATADIR;
 	if (dir == NULL)
 		return -ENOENT;
 
@@ -2298,9 +2350,7 @@ again:
 	while (spa_json_get_string(&it[1], key, sizeof(key)-1) > 0) {
 		bool add = false;
 
-		if (key[0] == '#') {
-			add = false;
-		} else if (pw_properties_get(impl->modules, key) != NULL) {
+		if (pw_properties_get(impl->modules, key) != NULL) {
 			add = true;
 		} else {
 			snprintf(check_path, sizeof(check_path),
@@ -2312,8 +2362,6 @@ again:
 				continue;
 
 			while (spa_json_get_string(&it[2], value, sizeof(value)-1) > 0) {
-				if (value[0] == '#')
-					continue;
 				pw_properties_set(impl->modules, value, "true");
 			}
 		}
@@ -2347,6 +2395,7 @@ static const struct {
 	{ "v4l2", "video for linux udev detection", sm_v4l2_monitor_start, NULL },
 	{ "libcamera", "libcamera udev detection", sm_libcamera_monitor_start, NULL },
 	{ "bluez5", "bluetooth support", sm_bluez5_monitor_start, NULL },
+	{ "bluez5-autoswitch", "switch bluetooth profiles automatically", sm_bluez5_autoswitch_start, NULL },
 	{ "suspend-node", "suspend inactive nodes", sm_suspend_node_start, NULL },
 	{ "policy-node", "configure and link nodes", sm_policy_node_start, NULL },
 	{ "pulse-bridge", "accept pulseaudio clients", sm_pulse_bridge_start, NULL },
@@ -2480,10 +2529,14 @@ int main(int argc, char *argv[])
 		impl.this.dbus_connection = spa_dbus_get_connection(impl.dbus, SPA_DBUS_TYPE_SESSION);
 		if (impl.this.dbus_connection == NULL)
 			pw_log_warn("no dbus connection");
-		else
+		else {
 			pw_log_debug("got dbus connection %p", impl.this.dbus_connection);
+			spa_dbus_connection_add_listener(impl.this.dbus_connection,
+					&impl.dbus_connection_listener,
+					&dbus_connection_events, &impl);
+		}
 	} else {
-		pw_log_info("dbus diabled");
+		pw_log_info("dbus disabled");
 	}
 
 	if ((res = start_session(&impl)) < 0)
