@@ -32,16 +32,20 @@
 
 #include <spa/node/node.h>
 #include <spa/utils/hook.h>
+#include <spa/utils/string.h>
 #include <spa/param/audio/format-utils.h>
 #include <spa/param/props.h>
 #include <spa/debug/pod.h>
 #include <spa/utils/json.h>
 
 #include "pipewire/pipewire.h"
-#include "extensions/metadata.h"
+#include "pipewire/extensions/metadata.h"
 
 #include "media-session.h"
 
+/** \page page_media_session_module_policy_node Media Session Module: Policy Node
+ */
+
 #define NAME		"policy-node"
 #define SESSION_KEY	"policy-node"
 
@@ -123,20 +127,22 @@ struct node {
 	unsigned int configured:1;
 	unsigned int dont_remix:1;
 	unsigned int monitor:1;
-	unsigned int moving:1;
 	unsigned int capture_sink:1;
 	unsigned int virtual:1;
 	unsigned int linking:1;
+	unsigned int have_passthrough:1;
+	unsigned int passthrough_only:1;
 };
 
-static int check_new_target(struct impl *impl, struct node *target);
-
 static bool find_format(struct node *node)
 {
 	struct impl *impl = node->impl;
 	struct sm_param *p;
 	bool have_format = false;
 
+	node->have_passthrough = false;
+	node->passthrough_only = false;
+
 	spa_list_for_each(p, &node->obj->param_list, link) {
 		struct spa_audio_info info = { 0, };
 		struct spa_pod *position = NULL;
@@ -148,37 +154,50 @@ static bool find_format(struct node *node)
 		if (spa_format_parse(p->param, &info.media_type, &info.media_subtype) < 0)
 			continue;
 
-		if (info.media_type != SPA_MEDIA_TYPE_audio ||
-		    info.media_subtype != SPA_MEDIA_SUBTYPE_raw)
+		if (info.media_type != SPA_MEDIA_TYPE_audio)
 			continue;
 
-		spa_pod_object_fixate((struct spa_pod_object*)p->param);
-		if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
-			spa_debug_pod(2, NULL, p->param);
+		switch (info.media_subtype) {
+		case SPA_MEDIA_SUBTYPE_raw:
+			spa_pod_object_fixate((struct spa_pod_object*)p->param);
+			if (pw_log_level_enabled(SPA_LOG_LEVEL_DEBUG))
+				spa_debug_pod(2, NULL, p->param);
 
-		/* defaults */
-		info.info.raw.format = SPA_AUDIO_FORMAT_F32;
-		info.info.raw.rate = impl->sample_rate;
-		info.info.raw.channels = 2;
-		info.info.raw.position[0] = SPA_AUDIO_CHANNEL_FL;
-		info.info.raw.position[1] = SPA_AUDIO_CHANNEL_FR;
+			/* defaults */
+			info.info.raw.format = SPA_AUDIO_FORMAT_F32;
+			info.info.raw.rate = impl->sample_rate;
+			info.info.raw.channels = 2;
+			info.info.raw.position[0] = SPA_AUDIO_CHANNEL_FL;
+			info.info.raw.position[1] = SPA_AUDIO_CHANNEL_FR;
 
-		spa_pod_parse_object(p->param,
-			SPA_TYPE_OBJECT_Format, NULL,
-			SPA_FORMAT_AUDIO_format,	SPA_POD_Id(&info.info.raw.format),
-			SPA_FORMAT_AUDIO_rate,		SPA_POD_OPT_Int(&info.info.raw.rate),
-			SPA_FORMAT_AUDIO_channels,	SPA_POD_Int(&info.info.raw.channels),
-			SPA_FORMAT_AUDIO_position,	SPA_POD_OPT_Pod(&position));
+			spa_pod_parse_object(p->param,
+				SPA_TYPE_OBJECT_Format, NULL,
+				SPA_FORMAT_AUDIO_format,	SPA_POD_Id(&info.info.raw.format),
+				SPA_FORMAT_AUDIO_rate,		SPA_POD_OPT_Int(&info.info.raw.rate),
+				SPA_FORMAT_AUDIO_channels,	SPA_POD_Int(&info.info.raw.channels),
+				SPA_FORMAT_AUDIO_position,	SPA_POD_OPT_Pod(&position));
 
-		if (position != NULL)
-			n_position = spa_pod_copy_array(position, SPA_TYPE_Id,
-					info.info.raw.position, SPA_AUDIO_MAX_CHANNELS);
-		if (n_position == 0 || n_position != info.info.raw.channels)
-			SPA_FLAG_SET(info.info.raw.flags, SPA_AUDIO_FLAG_UNPOSITIONED);
+			if (position != NULL)
+				n_position = spa_pod_copy_array(position, SPA_TYPE_Id,
+						info.info.raw.position, SPA_AUDIO_MAX_CHANNELS);
+			if (n_position == 0 || n_position != info.info.raw.channels)
+				SPA_FLAG_SET(info.info.raw.flags, SPA_AUDIO_FLAG_UNPOSITIONED);
 
-		if (node->format.info.raw.channels < info.info.raw.channels)
-			node->format = info;
+			if (node->format.info.raw.channels < info.info.raw.channels)
+				node->format = info;
 
+			have_format = true;
+			break;
+
+		case SPA_MEDIA_SUBTYPE_iec958:
+			pw_log_info("passthrough node %d found", node->id);
+			node->have_passthrough = true;
+			break;
+		}
+	}
+	if (!have_format && node->have_passthrough) {
+		pw_log_info("passthrough only node %d found", node->id);
+		node->passthrough_only = true;
 		have_format = true;
 	}
 	return have_format;
@@ -196,7 +215,7 @@ static int configure_node(struct node *node, struct spa_audio_info *info, bool f
 	if (node->configured && !force)
 		return 0;
 
-	if (strcmp(node->media, "Audio") != 0)
+	if (!spa_streq(node->media, "Audio"))
 		return 0;
 
 	format = node->format;
@@ -231,8 +250,10 @@ static int configure_node(struct node *node, struct spa_audio_info *info, bool f
 
 	node->configured = true;
 
-	if (node->type == NODE_TYPE_DEVICE)
-		check_new_target(impl, node);
+	if (node->type == NODE_TYPE_DEVICE) {
+		/* Schedule rescan in case we need to move streams */
+		sm_media_session_schedule_rescan(impl->session);
+	}
 
 	return 0;
 }
@@ -290,24 +311,24 @@ handle_node(struct impl *impl, struct sm_object *object)
 	spa_list_append(&impl->node_list, &node->link);
 	impl->node_list_changed = true;
 
-	if (role && !strcmp(role, "DSP"))
+	if (role && spa_streq(role, "DSP"))
 		node->active = node->configured = true;
 
-	if (strstr(media_class, "Stream/") == media_class) {
+	if (spa_strstartswith(media_class, "Stream/")) {
 		media_class += strlen("Stream/");
 
-		if (strstr(media_class, "Output/") == media_class) {
+		if (spa_strstartswith(media_class, "Output/")) {
 			direction = PW_DIRECTION_OUTPUT;
 			media_class += strlen("Output/");
 		}
-		else if (strstr(media_class, "Input/") == media_class) {
+		else if (spa_strstartswith(media_class, "Input/")) {
 			direction = PW_DIRECTION_INPUT;
 			media_class += strlen("Input/");
 		}
 		else
 			return 0;
 
-		if (strstr(media_class, "Video") == media_class) {
+		if (spa_strstartswith(media_class, "Video")) {
 			if (direction == PW_DIRECTION_OUTPUT) {
 				if ((str = pw_properties_get(object->props, PW_KEY_NODE_PLUGGED)) != NULL)
 					node->plugged = pw_properties_parse_uint64(str);
@@ -316,7 +337,7 @@ handle_node(struct impl *impl, struct sm_object *object)
 			}
 			node->active = node->configured = true;
 		}
-		else if (strstr(media_class, "Unknown") == media_class) {
+		else if (spa_strstartswith(media_class, "Unknown")) {
 			node->active = node->configured = true;
 		}
 
@@ -329,11 +350,11 @@ handle_node(struct impl *impl, struct sm_object *object)
 		const char *media;
 		bool virtual = false;
 
-		if (strstr(media_class, "Audio/") == media_class) {
+		if (spa_strstartswith(media_class, "Audio/")) {
 			media_class += strlen("Audio/");
 			media = "Audio";
 		}
-		else if (strstr(media_class, "Video/") == media_class) {
+		else if (spa_strstartswith(media_class, "Video/")) {
 			media_class += strlen("Video/");
 			media = "Video";
 			node->active = node->configured = true;
@@ -341,12 +362,12 @@ handle_node(struct impl *impl, struct sm_object *object)
 		else
 			return 0;
 
-		if (strcmp(media_class, "Sink") == 0 ||
-		    strcmp(media_class, "Duplex") == 0)
+		if (spa_streq(media_class, "Sink") ||
+		    spa_streq(media_class, "Duplex"))
 			direction = PW_DIRECTION_INPUT;
-		else if (strcmp(media_class, "Source") == 0)
+		else if (spa_streq(media_class, "Source"))
 			direction = PW_DIRECTION_OUTPUT;
-		else if (strcmp(media_class, "Source/Virtual") == 0) {
+		else if (spa_streq(media_class, "Source/Virtual")) {
 			virtual = true;
 			direction = PW_DIRECTION_OUTPUT;
 		} else
@@ -392,15 +413,6 @@ static void destroy_node(struct impl *impl, struct node *node)
 	sm_object_remove_data((struct sm_object*)node->obj, SESSION_KEY);
 }
 
-static inline int strzcmp(const char *s1, const char *s2)
-{
-	if (s1 == s2)
-		return 0;
-	if (s1 == NULL || s2 == NULL)
-		return 1;
-	return strcmp(s1, s2);
-}
-
 static int json_object_find(const char *obj, const char *key, char *value, size_t len)
 {
 	struct spa_json it[2];
@@ -412,7 +424,7 @@ static int json_object_find(const char *obj, const char *key, char *value, size_
 		return -EINVAL;
 
 	while (spa_json_get_string(&it[1], k, sizeof(k)-1) > 0) {
-		if (strcmp(k, key) == 0) {
+		if (spa_streq(k, key)) {
 			if (spa_json_get_string(&it[1], value, len) <= 0)
 				continue;
 			return 0;
@@ -428,7 +440,7 @@ static bool check_node_name(struct node *node, const char *name)
 {
 	const char *str;
 	if ((str = pw_properties_get(node->obj->obj.props, PW_KEY_NODE_NAME)) != NULL &&
-	    name != NULL && strcmp(str, name) == 0)
+	    name != NULL && spa_streq(str, name))
 		return true;
 	return false;
 }
@@ -447,6 +459,49 @@ static struct node *find_node_by_id_name(struct impl *impl, uint32_t id, const c
 	return NULL;
 }
 
+static bool can_link_check(struct impl *impl, const char *link_group, struct node *target, int hops)
+{
+	const char *g, *ng;
+	struct node *n;
+
+	if (hops == 8)
+		return false;
+
+	pw_log_debug("link group %s", link_group);
+
+	if (target->obj->info == NULL || target->obj->info->props == NULL)
+		return true;
+	g = spa_dict_lookup(target->obj->info->props, PW_KEY_NODE_LINK_GROUP);
+	if (g == NULL)
+		return true;
+	if (spa_streq(g, link_group))
+		return false;
+
+	spa_list_for_each(n, &impl->node_list, link) {
+		if (n == target || n->direction != target->direction)
+			continue;
+		if (n->obj->info == NULL || n->obj->info->props == NULL)
+			return true;
+		ng = spa_dict_lookup(n->obj->info->props, PW_KEY_NODE_LINK_GROUP);
+		if (ng == NULL || !spa_streq(ng, g))
+			continue;
+		if (n->peer != NULL && !can_link_check(impl, link_group, n->peer, hops + 1))
+			return false;
+	}
+	return true;
+}
+
+static bool can_link(struct impl *impl, struct node *node, struct node *target)
+{
+	const char *link_group;
+
+	link_group = spa_dict_lookup(node->obj->info->props, PW_KEY_NODE_LINK_GROUP);
+	if (link_group == NULL)
+		return true;
+
+	return can_link_check(impl, link_group, target, 0);
+}
+
 static const char *get_device_name(struct node *node)
 {
 	if (node->type != NODE_TYPE_DEVICE ||
@@ -466,7 +521,7 @@ static uint32_t find_device_for_name(struct impl *impl, const char *name)
 			return id;
 		if ((str = get_device_name(node)) == NULL)
 			continue;
-		if (strcmp(str, name) == 0)
+		if (spa_streq(str, name))
 			return node->obj->obj.id;
 	}
 	return SPA_ID_INVALID;
@@ -479,7 +534,7 @@ static void session_create(void *data, struct sm_object *object)
 
 	clock_gettime(CLOCK_MONOTONIC, &impl->now);
 
-	if (strcmp(object->type, PW_TYPE_INTERFACE_Node) == 0)
+	if (spa_streq(object->type, PW_TYPE_INTERFACE_Node))
 		res = handle_node(impl, object);
 	else
 		res = 0;
@@ -495,7 +550,7 @@ static void session_remove(void *data, struct sm_object *object)
 	struct impl *impl = data;
 	pw_log_debug(NAME " %p: remove global '%d'", impl, object->id);
 
-	if (strcmp(object->type, PW_TYPE_INTERFACE_Node) == 0) {
+	if (spa_streq(object->type, PW_TYPE_INTERFACE_Node)) {
 		struct node *n, *node;
 
 		if ((node = sm_object_get_data(object, SESSION_KEY)) != NULL)
@@ -516,6 +571,7 @@ struct find_data {
 	struct node *node;
 
 	const char *media;
+	const char *link_group;
 	bool capture_sink;
 	enum pw_direction direction;
 
@@ -549,15 +605,20 @@ static int find_node(void *data, struct node *node)
 		return 0;
 	}
 
-	if (node->media && strcmp(node->media, find->media) != 0) {
+	if (node->media && !spa_streq(node->media, find->media)) {
 		pw_log_debug(".. incompatible media %s <-> %s", node->media, find->media);
 		return 0;
 	}
+	if (find->link_group && !can_link_check(impl, find->link_group, node, 0)) {
+		pw_log_info(".. connecting link-group %s", find->link_group);
+		return 0;
+	}
+
 	plugged = node->plugged;
 	priority = node->priority;
 
 	if (node->media) {
-		if (strcmp(node->media, "Audio") == 0) {
+		if (spa_streq(node->media, "Audio")) {
 			if (node->direction == PW_DIRECTION_INPUT) {
 				if (find->direction == PW_DIRECTION_OUTPUT)
 					is_default |= check_node_name(node,
@@ -569,7 +630,7 @@ static int find_node(void *data, struct node *node)
 			    find->direction == PW_DIRECTION_INPUT)
 				is_default |= check_node_name(node,
 						impl->defaults[DEFAULT_AUDIO_SOURCE].config);
-		} else if (strcmp(node->media, "Video") == 0) {
+		} else if (spa_streq(node->media, "Video")) {
 			if (node->direction == PW_DIRECTION_OUTPUT &&
 			    find->direction == PW_DIRECTION_INPUT)
 				is_default |= check_node_name(node,
@@ -614,13 +675,13 @@ static struct node *find_auto_default_node(struct impl *impl, const struct defau
 	find.capture_sink = false;
 	find.exclusive = false;
 
-	if (strcmp(def->key, DEFAULT_AUDIO_SINK_KEY) == 0) {
+	if (spa_streq(def->key, DEFAULT_AUDIO_SINK_KEY)) {
 		find.media = "Audio";
 		find.direction = PW_DIRECTION_OUTPUT;
-	} else if (strcmp(def->key, DEFAULT_AUDIO_SOURCE_KEY) == 0) {
+	} else if (spa_streq(def->key, DEFAULT_AUDIO_SOURCE_KEY)) {
 		find.media = "Audio";
 		find.direction = PW_DIRECTION_INPUT;
-	} else if (strcmp(def->key, DEFAULT_VIDEO_SOURCE_KEY) == 0) {
+	} else if (spa_streq(def->key, DEFAULT_VIDEO_SOURCE_KEY)) {
 		find.media = "Video";
 		find.direction = PW_DIRECTION_INPUT;
 	} else {
@@ -673,14 +734,7 @@ static int link_nodes(struct node *node, struct node *peer)
 
 	if (res > 0) {
 		node->peer = peer;
-		node->failed_peer = NULL;
 		node->connect_count++;
-		node->failed_count = 0;
-	} else {
-		if (node->failed_peer != peer)
-			node->failed_count = 0;
-		node->failed_peer = peer;
-		node->failed_count++;
 	}
 	return res;
 }
@@ -713,26 +767,56 @@ static int unlink_nodes(struct node *node, struct node *peer)
 	return 0;
 }
 
+static int relink_node(struct impl *impl, struct node *n, struct node *peer)
+{
+	int res;
+
+	if (peer == n->failed_peer && n->failed_count > MAX_LINK_RETRY) {
+		/* Break rescan -> failed link -> rescan loop. */
+		pw_log_debug(NAME" %p: tried to link '%d' on last rescan, not retrying",
+				impl, peer->id);
+		return -EBUSY;
+	}
+
+	if (n->failed_peer != peer)
+		n->failed_count = 0;
+	n->failed_peer = peer;
+	n->failed_count++;
+
+	if (!can_link(impl, n, peer)) {
+		pw_log_debug("can't link node %d to %d: same link-group", n->id,
+				peer->id);
+		return -EPERM;
+	}
+
+	if (n->peer != NULL)
+		if ((res = unlink_nodes(n, n->peer)) < 0)
+			return res;
+
+	pw_log_debug(NAME" %p: linking node %d to node %d", impl, n->id, peer->id);
+
+	/* NB. if link_nodes returns error, n may have been invalidated */
+	if ((res = link_nodes(n, peer)) > 0) {
+		n->failed_peer = NULL;
+		n->failed_count = 0;
+	}
+	return res;
+}
+
 static int rescan_node(struct impl *impl, struct node *n)
 {
 	struct spa_dict *props;
 	const char *str;
 	bool exclusive, reconnect, autoconnect;
-	struct find_data find;
 	struct pw_node_info *info;
 	struct node *peer;
 	struct sm_object *obj;
 	uint32_t path_id;
-	bool follows_default;
 
 	if (!n->active) {
 		pw_log_debug(NAME " %p: node %d is not active", impl, n->id);
 		return 0;
 	}
-	if (n->moving) {
-		pw_log_debug(NAME " %p: node %d is moving", impl, n->id);
-		return 0;
-	}
 
 	if (n->type == NODE_TYPE_DEVICE) {
 		configure_node(n, NULL, false);
@@ -750,18 +834,6 @@ static int rescan_node(struct impl *impl, struct node *n)
 	str = spa_dict_lookup(props, PW_KEY_NODE_DONT_RECONNECT);
 	reconnect = str ? !pw_properties_parse_bool(str) : true;
 
-	follows_default = (impl->streams_follow_default &&
-	                   n->type == NODE_TYPE_STREAM &&
-	                   reconnect &&
-	                   n->obj->target_node == NULL &&
-	                   ((str = spa_dict_lookup(props, PW_KEY_NODE_TARGET)) == NULL ||
-			    (uint32_t)atoi(str) == SPA_ID_INVALID));
-
-	if (n->peer != NULL && !follows_default) {
-		pw_log_debug(NAME " %p: node %d is already linked", impl, n->id);
-		return 0;
-	}
-
 	if ((str = spa_dict_lookup(props, PW_KEY_STREAM_DONT_REMIX)) != NULL)
 		n->dont_remix = pw_properties_parse_bool(str);
 
@@ -777,7 +849,7 @@ static int rescan_node(struct impl *impl, struct node *n)
 		autoconnect = pw_properties_parse_bool(str);
 
 	if ((str = spa_dict_lookup(props, PW_KEY_DEVICE_API)) != NULL &&
-	    strcmp(str, "bluez5") == 0)
+	    spa_streq(str, "bluez5"))
 		autoconnect = true;
 
 	if (!autoconnect) {
@@ -793,40 +865,49 @@ static int rescan_node(struct impl *impl, struct node *n)
 
 	str = spa_dict_lookup(props, PW_KEY_NODE_EXCLUSIVE);
 	exclusive = str ? pw_properties_parse_bool(str) : false;
-
 	pw_log_debug(NAME " %p: exclusive:%d", impl, exclusive);
 
-	spa_zero(find);
-	find.impl = impl;
-	find.media = n->media;
-	find.capture_sink = n->capture_sink;
-	find.direction = n->direction;
-	find.exclusive = exclusive;
-
-	/* we always honour the target node asked for by the client */
+	/* honor target node set by user or asked for by the client */
 	path_id = SPA_ID_INVALID;
-	if ((str = spa_dict_lookup(props, PW_KEY_NODE_TARGET)) != NULL)
-		path_id = find_device_for_name(impl, str);
-	if (path_id == SPA_ID_INVALID && n->obj->target_node != NULL)
+	if (n->obj->target_node != NULL)
 		path_id = find_device_for_name(impl, n->obj->target_node);
-
-	pw_log_info("trying to link node %d exclusive:%d reconnect:%d target:%d follows-default:%d", n->id,
-	            exclusive, reconnect, path_id, follows_default);
+	if (!n->obj->fixed_target &&
+			(str = spa_dict_lookup(props, PW_KEY_NODE_TARGET)) != NULL) {
+		bool has_target = ((uint32_t)atoi(str) != SPA_ID_INVALID);
+		path_id = find_device_for_name(impl, str);
+		if (!reconnect && has_target && path_id == SPA_ID_INVALID) {
+			/* don't use fallbacks for non-reconnecting nodes */
+			peer = NULL;
+			goto do_link;
+		}
+	}
 
 	if (n->peer != NULL) {
-		spa_list_for_each(peer, &impl->node_list, link)
-			find_node(&find, peer);
-
-		if (follows_default && find.node != NULL && find.node != n->peer) {
-			pw_log_debug(NAME " %p: node %d follows default, changed (%d -> %d)", impl, n->id,
-			             n->peer->id, find.node->id);
-			unlink_nodes(n, n->peer);
-		} else {
-			pw_log_debug(NAME " %p: node %d already linked (not changing)", impl, n->id);
+		/* Do we need to check again where to link to? */
+		bool target_found = (path_id != SPA_ID_INVALID);
+		bool peer_is_target = (target_found && n->peer->obj->obj.id == path_id);
+		bool follows_default = (impl->streams_follow_default &&
+				n->type == NODE_TYPE_STREAM);
+		bool recheck = !peer_is_target && (follows_default || target_found) &&
+			reconnect;
+		if (!recheck) {
+			pw_log_debug(NAME " %p: node %d is already linked, peer-is-target:%d "
+					"follows-default:%d", impl, n->id, peer_is_target,
+					follows_default);
 			return 0;
 		}
 	}
 
+	if (n->passthrough_only) {
+		path_id = SPA_ID_INVALID;
+		peer = NULL;
+		reconnect = false;
+		goto do_link;
+	}
+
+	pw_log_info("trying to link node %d exclusive:%d reconnect:%d target:%d",
+			n->id, exclusive, reconnect, path_id);
+
 	if (path_id != SPA_ID_INVALID) {
 		pw_log_debug(NAME " %p: target:%d", impl, path_id);
 
@@ -836,10 +917,8 @@ static int rescan_node(struct impl *impl, struct node *n)
 		if ((obj = sm_media_session_find_object(impl->session, path_id)) != NULL) {
 			pw_log_debug(NAME " %p: found target:%d type:%s", impl,
 					path_id, obj->type);
-			if (strcmp(obj->type, PW_TYPE_INTERFACE_Node) == 0) {
+			if (spa_streq(obj->type, PW_TYPE_INTERFACE_Node)) {
 				peer = sm_object_get_data(obj, SESSION_KEY);
-				if (peer == NULL)
-					return -ENOENT;
 				goto do_link;
 			}
 		}
@@ -847,34 +926,54 @@ static int rescan_node(struct impl *impl, struct node *n)
 				path_id, reconnect);
 	}
 	if (path_id == SPA_ID_INVALID && (reconnect || n->connect_count == 0)) {
-		if (find.node == NULL)
-			spa_list_for_each(peer, &impl->node_list, link)
-				find_node(&find, peer);
+		/* find fallback */
+		struct find_data find;
+
+		spa_zero(find);
+		find.impl = impl;
+		find.media = n->media;
+		find.capture_sink = n->capture_sink;
+		find.direction = n->direction;
+		find.exclusive = exclusive;
+		find.link_group = n->peer == NULL ? spa_dict_lookup(props, PW_KEY_NODE_LINK_GROUP) : NULL;
+
+		spa_list_for_each(peer, &impl->node_list, link)
+			find_node(&find, peer);
+
+		peer = find.node;
 	} else {
-		find.node = NULL;
+		peer = NULL;
 	}
 
-	if (find.node == NULL) {
-		struct sm_object *obj;
-
+do_link:
+	if (peer == NULL) {
 		if (!reconnect) {
 			pw_log_info("don-reconnect target node destroyed: destroy %d", n->id);
 			sm_media_session_destroy_object(impl->session, n->id);
+		} else if (reconnect && n->connect_count > 0) {
+			/* Don't error the stream on reconnects */
+			pw_log_info(NAME " %p: no node found for %d, waiting reconnect", impl, n->id);
+			if (n->peer != NULL)
+				unlink_nodes(n, n->peer);
+			return 0;
 		} else {
-			pw_log_warn("no node found for %d", n->id);
+			pw_log_warn(NAME " %p: no node found for %d, stream error", impl, n->id);
 		}
 
 		obj = sm_media_session_find_object(impl->session, n->client_id);
 		pw_log_debug(NAME " %p: client_id:%d object:%p type:%s", impl,
 				n->client_id, obj, obj ? obj->type : "None");
 
-		if (obj && strcmp(obj->type, PW_TYPE_INTERFACE_Client) == 0) {
+		if (obj && spa_streq(obj->type, PW_TYPE_INTERFACE_Client)) {
 			pw_client_error((struct pw_client*)obj->proxy,
 				n->id, -ENOENT, "no node available");
 		}
 		return -ENOENT;
+	} else if (peer == n->peer) {
+		pw_log_debug(NAME " %p: node %d already linked to %d (not changing)",
+				impl, n->id, peer->id);
+		return 0;
 	}
-	peer = find.node;
 
 	if (exclusive && peer->obj->info->state == PW_NODE_STATE_RUNNING) {
 		pw_log_warn("node %d busy, can't get exclusive access", peer->id);
@@ -882,17 +981,7 @@ static int rescan_node(struct impl *impl, struct node *n)
 	}
 	n->exclusive = exclusive;
 
-	pw_log_debug(NAME" %p: linking to node '%d'", impl, peer->id);
-
-do_link:
-	if (peer == n->failed_peer && n->failed_count > MAX_LINK_RETRY) {
-		/* Break rescan -> failed link -> rescan loop. */
-		pw_log_debug(NAME" %p: tried to link '%d' on last rescan, not retrying",
-				impl, peer->id);
-		return 0;
-	}
-	link_nodes(n, peer);
-	return 1;
+	return relink_node(impl, n, peer);
 }
 
 static void session_info(void *data, const struct pw_core_info *info)
@@ -930,7 +1019,7 @@ static void refresh_auto_default_nodes(struct impl *impl)
 			const char *name = pw_properties_get(node->obj->obj.props, PW_KEY_NODE_NAME);
 			char buf[1024];
 
-			if (name == NULL || strzcmp(name, def->value) == 0)
+			if (name == NULL || spa_streq(name, def->value))
 				continue;
 
 			free(def->value);
@@ -985,63 +1074,6 @@ static const struct sm_media_session_events session_events = {
 	.destroy = session_destroy,
 };
 
-static int do_move_node(struct node *n, struct node *src, struct node *dst)
-{
-	n->moving = true;
-	if (src)
-		unlink_nodes(n, src);
-	if (dst)
-		link_nodes(n, dst);
-	n->moving = false;
-	return 0;
-}
-
-static int handle_move(struct impl *impl, struct node *src_node, struct node *dst_node)
-{
-	const char *str;
-	struct pw_node_info *info;
-
-	if (src_node->peer == dst_node)
-		return 0;
-
-	if ((info = src_node->obj->info) == NULL)
-		return -EIO;
-
-	if ((str = spa_dict_lookup(info->props, PW_KEY_NODE_DONT_RECONNECT)) != NULL &&
-		    pw_properties_parse_bool(str)) {
-		pw_log_warn("can't reconnect node %d to %d", src_node->id,
-				dst_node->id);
-		return -EPERM;
-	}
-
-	pw_log_info("move node %d: from peer %d to %d", src_node->id,
-			src_node->peer ? src_node->peer->id : SPA_ID_INVALID,
-			dst_node->id);
-
-	free(src_node->obj->target_node);
-	str = get_device_name(dst_node);
-	src_node->obj->target_node = str ? strdup(str) : NULL;
-
-	return do_move_node(src_node, src_node->peer, dst_node);
-}
-
-static int check_new_target(struct impl *impl, struct node *target)
-{
-	struct node *node;
-	const char *str = get_device_name(target);
-
-	spa_list_for_each(node, &impl->node_list, link) {
-		pw_log_debug(NAME" %p: node %d target '%s' find:%s", impl,
-				node->id, node->obj->target_node, str);
-
-		if (node->obj->target_node != NULL &&
-		    strcmp(node->obj->target_node , str) == 0) {
-			handle_move(impl, node, target);
-		}
-	}
-	return 0;
-}
-
 static int metadata_property(void *object, uint32_t subject,
 		const char *key, const char *type, const char *value)
 {
@@ -1061,14 +1093,14 @@ static int metadata_property(void *object, uint32_t subject,
 			val = name;
 		}
 		for (def = impl->defaults; def->key != NULL; ++def) {
-			if (key == NULL || strcmp(key, def->key_config) == 0) {
-				if (strzcmp(def->config, val) != 0)
+			if (key == NULL || spa_streq(key, def->key_config)) {
+				if (!spa_streq(def->config, val))
 					changed = true;
 				free(def->config);
 				def->config = val ? strdup(val) : NULL;
 			}
-			if (key == NULL || strcmp(key, def->key) == 0) {
-				bool eff_changed = strzcmp(def->value, val) != 0;
+			if (key == NULL || spa_streq(key, def->key)) {
+				bool eff_changed = !spa_streq(def->value, val);
 				free(def->value);
 				def->value = val ? strdup(val) : NULL;
 
@@ -1080,25 +1112,39 @@ static int metadata_property(void *object, uint32_t subject,
 		}
 		if (changed)
 			sm_media_session_schedule_rescan(impl->session);
-	} else if (key != NULL && strcmp(key, "target.node") == 0) {
-		if (value != NULL) {
-			struct node *src_node, *dst_node;
+	} else if (key == NULL || spa_streq(key, "target.node")) {
+		struct node *src_node;
+
+		src_node = find_node_by_id_name(impl, subject, NULL);
+		if (!src_node)
+			return 0;
+
+		/* Set target and schedule rescan */
+		if (key == NULL || value == NULL) {
+			free(src_node->obj->target_node);
+			src_node->obj->target_node = NULL;
+			src_node->obj->fixed_target = false;
+		} else {
+			const char *str;
+			struct node *dst_node;
 
 			dst_node = find_node_by_id_name(impl, SPA_ID_INVALID, value);
-			src_node = dst_node ? find_node_by_id_name(impl, subject, NULL) : NULL;
-
-			if (dst_node && src_node)
-				handle_move(impl, src_node, dst_node);
-		} else {
-			/* Unset target node. Schedule rescan to re-link, if needed. */
-			struct node *src_node;
-			src_node = find_node_by_id_name(impl, subject, NULL);
-			if (src_node) {
-				free(src_node->obj->target_node);
-				src_node->obj->target_node = NULL;
-				sm_media_session_schedule_rescan(impl->session);
+			if (dst_node) {
+				str = get_device_name(dst_node);
+				if (!str)
+					return 0;
+			} else if ((uint32_t)atoi(value) == SPA_ID_INVALID) {
+				str = NULL;
+			} else {
+				return 0;
 			}
+
+			free(src_node->obj->target_node);
+			src_node->obj->target_node = str ? strdup(str) : NULL;
+			src_node->obj->fixed_target = true;
 		}
+
+		sm_media_session_schedule_rescan(impl->session);
 	}
 	return 0;
 }
