@@ -32,6 +32,7 @@
 
 #include <spa/node/node.h>
 #include <spa/utils/hook.h>
+#include <spa/utils/string.h>
 #include <spa/param/props.h>
 #include <spa/debug/pod.h>
 
@@ -39,6 +40,9 @@
 
 #include "media-session.h"
 
+/** \page page_media_session_module_suspend_node Media Session Module: Suspend Node
+ */
+
 #define NAME		"suspend-node"
 #define SESSION_KEY	"suspend-node"
 
@@ -87,7 +91,7 @@ static void idle_timeout(void *data, uint64_t expirations)
 	struct impl *impl = node->impl;
 	struct spa_command *cmd = &SPA_NODE_COMMAND_INIT(SPA_NODE_COMMAND_Suspend);
 
-	pw_log_debug(NAME " %p: node %d idle timeout", impl, node->id);
+	pw_log_info(NAME " %p: node %d suspend", impl, node->id);
 
 	remove_idle_timeout(node);
 
@@ -177,8 +181,8 @@ handle_node(struct impl *impl, struct sm_object *object)
 	if (media_class == NULL)
 		return 0;
 
-	if (strstr(media_class, "Audio/") != media_class &&
-	    (strstr(media_class, "Video/") != media_class))
+	if (!spa_strstartswith(media_class, "Audio/") &&
+	    (!spa_strstartswith(media_class, "Video/")))
 		return 0;
 
 	node = sm_object_add_data(object, SESSION_KEY, sizeof(struct node));
@@ -206,7 +210,7 @@ static void session_create(void *data, struct sm_object *object)
 	struct impl *impl = data;
 	int res;
 
-	if (strcmp(object->type, PW_TYPE_INTERFACE_Node) == 0)
+	if (spa_streq(object->type, PW_TYPE_INTERFACE_Node))
 		res = handle_node(impl, object);
 	else
 		res = 0;
@@ -220,7 +224,7 @@ static void session_remove(void *data, struct sm_object *object)
 	struct impl *impl = data;
 	pw_log_debug(NAME " %p: remove global '%d'", impl, object->id);
 
-	if (strcmp(object->type, PW_TYPE_INTERFACE_Node) == 0) {
+	if (spa_streq(object->type, PW_TYPE_INTERFACE_Node)) {
 		struct node *node;
 		if ((node = sm_object_get_data(object, SESSION_KEY)) != NULL)
 			destroy_node(impl, node);
