@@ -36,6 +36,7 @@ extern "C" {
 
 #include <spa/support/plugin.h>
 #include <spa/pod/builder.h>
+#include <spa/param/latency-utils.h>
 #include <spa/utils/result.h>
 #include <spa/utils/type-info.h>
 
@@ -48,17 +49,29 @@ struct ucred {
 #define spa_debug(...) pw_log_trace(__VA_ARGS__)
 #endif
 
-struct defaults {
-	uint32_t clock_rate;
-	uint32_t clock_quantum;
-	uint32_t clock_min_quantum;
-	uint32_t clock_max_quantum;
+#define MAX_RATES				16u
+#define CLOCK_MIN_QUANTUM			4u
+#define CLOCK_MAX_QUANTUM			8192u
+
+struct settings {
+	uint32_t log_level;
+	uint32_t clock_rate;			/* default clock rate */
+	uint32_t clock_rates[MAX_RATES];	/* allowed clock rates */
+	uint32_t n_clock_rates;			/* number of alternative clock rates */
+	uint32_t clock_quantum;			/* default quantum */
+	uint32_t clock_min_quantum;		/* min quantum */
+	uint32_t clock_max_quantum;		/* max quantum */
 	struct spa_rectangle video_size;
 	struct spa_fraction video_rate;
 	uint32_t link_max_buffers;
 	unsigned int mem_warn_mlock:1;
 	unsigned int mem_allow_mlock:1;
 	unsigned int clock_power_of_two_quantum:1;
+#define CLOCK_RATE_UPDATE_MODE_HARD 0
+#define CLOCK_RATE_UPDATE_MODE_SOFT 1
+	int clock_rate_update_mode;
+	uint32_t clock_force_rate;		/* force a clock rate */
+	uint32_t clock_force_quantum;		/* force a quantum */
 };
 
 struct ratelimit {
@@ -68,11 +81,11 @@ struct ratelimit {
 	unsigned n_printed, n_missed;
 };
 
-static inline bool ratelimit_test(struct ratelimit *r, uint64_t now)
+static inline bool ratelimit_test(struct ratelimit *r, uint64_t now, enum spa_log_level level)
 {
 	if (r->begin + r->interval < now) {
 		if (r->n_missed)
-			pw_log_warn("%u events suppressed", r->n_missed);
+			pw_log(level, "%u events suppressed", r->n_missed);
 		r->begin = now;
 		r->n_printed = 0;
 		r->n_missed = 0;
@@ -125,7 +138,7 @@ static inline struct pw_param *pw_param_add(struct spa_list *params,
 
 	p->id = id;
 	if (param != NULL) {
-		p->param = SPA_MEMBER(p, sizeof(*p), struct spa_pod);
+		p->param = SPA_PTROFF(p, sizeof(*p), struct spa_pod);
 		memcpy(p->param, param, SPA_POD_SIZE(param));
 	} else {
 		pw_param_clear(params, id);
@@ -174,7 +187,7 @@ struct pw_protocol {
 	struct spa_list server_list;            /**< list of current servers */
 	struct spa_hook_list listener_list;	/**< event listeners */
 
-	const struct pw_protocol_implementaton *implementation; /**< implementation of the protocol */
+	const struct pw_protocol_implementation *implementation; /**< implementation of the protocol */
 
 	const void *extension;  /**< extension API */
 
@@ -234,6 +247,29 @@ struct pw_impl_core {
 	unsigned int registered:1;
 };
 
+#define pw_impl_metadata_emit(s,m,v,...) spa_hook_list_call(&s->listener_list, struct pw_impl_metadata_events, m, v, ##__VA_ARGS__)
+
+#define pw_impl_metadata_emit_destroy(s)	pw_impl_metadata_emit(s, destroy, 0)
+#define pw_impl_metadata_emit_free(s)		pw_impl_metadata_emit(s, free, 0)
+#define pw_impl_metadata_emit_property(s, ...)	pw_impl_metadata_emit(s, property, 0, __VA_ARGS__)
+
+struct pw_impl_metadata {
+	struct pw_context *context;		/**< the context */
+	struct spa_list link;			/**< link in context metadata_list */
+	struct pw_global *global;		/**< global for this metadata */
+	struct spa_hook global_listener;
+
+	struct pw_properties *properties;	/**< properties of the metadata */
+
+	struct pw_metadata *metadata;
+	struct spa_hook metadata_listener;
+
+	struct spa_hook_list listener_list;	/**< event listeners */
+	void *user_data;
+
+	unsigned int registered:1;
+};
+
 struct pw_impl_client {
 	struct pw_impl_core *core;		/**< core object */
 	struct pw_context *context;		/**< context object */
@@ -377,7 +413,10 @@ struct pw_context {
 	struct pw_properties *conf;		/**< configuration of the context */
 	struct pw_properties *properties;	/**< properties of the context */
 
-	struct defaults defaults;		/**< default parameters */
+	struct settings defaults;		/**< default parameters */
+	struct settings settings;		/**< current parameters */
+
+	void *settings_impl;		/**< settings metadata */
 
 	struct pw_mempool *pool;		/**< global memory pool */
 
@@ -393,6 +432,7 @@ struct pw_context {
 	struct spa_list client_list;		/**< list of clients */
 	struct spa_list node_list;		/**< list of nodes */
 	struct spa_list factory_list;		/**< list of factories */
+	struct spa_list metadata_list;		/**< list of metadata */
 	struct spa_list link_list;		/**< list of links */
 	struct spa_list control_list[2];	/**< list of controls, indexed by direction */
 	struct spa_list export_list;		/**< list of export types */
@@ -416,6 +456,7 @@ struct pw_context {
 	struct pw_impl_client *current_client;	/**< client currently executing code in mainloop */
 
 	long sc_pagesize;
+	unsigned int freewheeling:1;
 
 	void *user_data;		/**< extra user data */
 };
@@ -622,7 +663,7 @@ struct pw_impl_node {
 	char *name;				/** for debug */
 
 	uint32_t priority_driver;	/** priority for being driver */
-	uint32_t group_id;		/** group to schedule this node in */
+	char group[128];		/** group to schedule this node in */
 	uint64_t spa_flags;
 
 	unsigned int registered:1;
@@ -636,6 +677,11 @@ struct pw_impl_node {
 	unsigned int visited:1;		/**< for sorting */
 	unsigned int want_driver:1;	/**< this node wants to be assigned to a driver */
 	unsigned int passive:1;		/**< driver graph only has passive links */
+	unsigned int freewheel:1;	/**< if this is the freewheel driver */
+	unsigned int loopchecked:1;	/**< for feedback loop checking */
+	unsigned int always_process:1;	/**< this node wants to always be processing, even when idle */
+	unsigned int lock_quantum:1;	/**< don't change graph quantum */
+	unsigned int lock_rate:1;	/**< don't change graph rate */
 
 	uint32_t port_user_data_size;	/**< extra size for port user data */
 
@@ -659,9 +705,8 @@ struct pw_impl_node {
 	struct pw_loop *data_loop;		/**< the data loop for this node */
 
 	struct spa_fraction latency;		/**< requested latency */
-	uint32_t quantum_size;			/**< desired quantum */
-	struct spa_fraction max_latency;	/**< miximum latency */
-	uint32_t max_quantum_size;		/**< max supported quantum */
+	struct spa_fraction max_latency;	/**< maximum latency */
+	struct spa_fraction rate;		/**< requested rate */
 	struct spa_source source;		/**< source to remotely trigger this node */
 	struct pw_memblock *activation;
 	struct {
@@ -695,6 +740,7 @@ struct pw_impl_port_mix {
 	} port;
 	struct spa_io_buffers *io;
 	uint32_t id;
+	uint32_t peer_id;
 	unsigned int have_buffers:1;
 };
 
@@ -729,6 +775,7 @@ struct pw_impl_port_implementation {
 #define pw_impl_port_emit_control_added(p,c)		pw_impl_port_emit(p, control_added, 0, c)
 #define pw_impl_port_emit_control_removed(p,c)		pw_impl_port_emit(p, control_removed, 0, c)
 #define pw_impl_port_emit_param_changed(p,i)		pw_impl_port_emit(p, param_changed, 1, i)
+#define pw_impl_port_emit_latency_changed(p)		pw_impl_port_emit(p, latency_changed, 2)
 
 #define PW_IMPL_PORT_IS_CONTROL(port)	SPA_FLAG_MASK(port->flags, \
 						PW_IMPL_PORT_FLAG_BUFFERS|PW_IMPL_PORT_FLAG_CONTROL,\
@@ -788,6 +835,10 @@ struct pw_impl_port {
 		struct spa_list node_link;
 	} rt;					/**< data only accessed from the data thread */
 	unsigned int added:1;
+	unsigned int destroying:1;
+
+	struct spa_latency_info latency[2];	/**< latencies */
+	unsigned int have_latency_param:1;
 
 	void *owner_data;		/**< extra owner data */
 	void *user_data;                /**< extra user data */
@@ -1017,7 +1068,7 @@ struct pw_filter {
 
 struct pw_impl_factory {
 	struct pw_context *context;		/**< the context */
-	struct spa_list link;		/**< link in context node_factory_list */
+	struct spa_list link;		/**< link in context factory_list */
 	struct pw_global *global;	/**< global for this factory */
 	struct spa_hook global_listener;
 
@@ -1105,10 +1156,10 @@ int pw_impl_port_release_mix(struct pw_impl_port *port, struct pw_impl_port_mix
 
 void pw_impl_port_update_state(struct pw_impl_port *port, enum pw_impl_port_state state, int res, char *error);
 
-/** Unlink a port \memberof pw_impl_port */
+/** Unlink a port */
 void pw_impl_port_unlink(struct pw_impl_port *port);
 
-/** Destroy a port \memberof pw_impl_port */
+/** Destroy a port */
 void pw_impl_port_destroy(struct pw_impl_port *port);
 
 /** Iterate the params of the given port. The callback should return
@@ -1142,15 +1193,17 @@ int pw_impl_port_for_each_link(struct pw_impl_port *port,
 			   int (*callback) (void *data, struct pw_impl_link *link),
 			   void *data);
 
-/** Set a param on a port \memberof pw_impl_port, use SPA_ID_INVALID for mix_id to set
+/** Set a param on a port, use SPA_ID_INVALID for mix_id to set
  * the param on all mix ports */
 int pw_impl_port_set_param(struct pw_impl_port *port,
 		uint32_t id, uint32_t flags, const struct spa_pod *param);
 
-/** Use buffers on a port \memberof pw_impl_port */
+/** Use buffers on a port */
 int pw_impl_port_use_buffers(struct pw_impl_port *port, struct pw_impl_port_mix *mix, uint32_t flags,
 		struct spa_buffer **buffers, uint32_t n_buffers);
 
+int pw_impl_port_recalc_latency(struct pw_impl_port *port);
+
 /** Change the state of the node */
 int pw_impl_node_set_state(struct pw_impl_node *node, enum pw_node_state state);
 
@@ -1161,13 +1214,13 @@ int pw_impl_node_update_ports(struct pw_impl_node *node);
 
 int pw_impl_node_set_driver(struct pw_impl_node *node, struct pw_impl_node *driver);
 
-/** Prepare a link \memberof pw_impl_link
+/** Prepare a link
   * Starts the negotiation of formats and buffers on \a link */
 int pw_impl_link_prepare(struct pw_impl_link *link);
 /** starts streaming on a link */
 int pw_impl_link_activate(struct pw_impl_link *link);
 
-/** Deactivate a link \memberof pw_impl_link */
+/** Deactivate a link */
 int pw_impl_link_deactivate(struct pw_impl_link *link);
 
 struct pw_control *
@@ -1202,6 +1255,9 @@ void pw_log_log_object(enum spa_log_level level, const char *file, int line,
 
 bool pw_log_is_default(void);
 
+int pw_settings_init(struct pw_context *context);
+void pw_settings_clean(struct pw_context *context);
+
 /** \endcond */
 
 #ifdef __cplusplus
